#
#   dialog.py
#
#   these are routine to show dialog
#   these are made to eliminate import wx
#   from mto fils.
#
#   read  : file read dialog
#   write : file write dialog
#   dlg_message   :  message dialog
#   dlg_textentry  : text entry dialog
#
import wx
import os
from ifigure.utils.wx3to4 import TextEntryDialog, deref_proxy


def read(parent=None,
         message='Select file to read',
         wildcard='*',
         defaultfile='',
         defaultdir=''):

    open_dlg = wx.FileDialog(parent, message=message,
                             wildcard=wildcard, style=wx.FD_OPEN)
    if defaultfile != '':
        open_dlg.SetFilename(os.path.basename(defaultfile))
        if os.path.dirname(defaultfile) != '':
            open_dlg.SetDirectory(os.path.dirname(defaultfile))
    if defaultdir != '':
        open_dlg.SetDirectory(defaultdir)

    path = ''
    if open_dlg.ShowModal() == wx.ID_OK:
        path = open_dlg.GetPath()
        open_dlg.Destroy()
    open_dlg.Destroy()
    return path


def readdir(parent=None, message='Select directory to read', wildcard='*', defaultfile=''):

    open_dlg = wx.DirDialog(parent, message=message,)
    if defaultfile != '':
        open_dlg.SetFilename(os.path.basename(defaultfile))
        if os.path.dirname(defaultfile) != '':
            open_dlg.SetDirectory(os.path.dirname(defaultfile))
    path = ''
    if open_dlg.ShowModal() == wx.ID_OK:
        path = open_dlg.GetPath()
        open_dlg.Destroy()
    open_dlg.Destroy()
    return path


def write(parent=None, defaultfile='',
          message='Select file to write', wildcard='*',
          return_filterindex=False,
          warn_overwrite=False):
    '''
    wrap FileDialog, Note that defaultfile can be absolute path
    '''
    style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT if warn_overwrite else wx.FD_SAVE

    open_dlg = wx.FileDialog(parent, message=message,
                             wildcard=wildcard,
                             style=style)
    if defaultfile != '':
        open_dlg.SetFilename(os.path.basename(defaultfile))
        if os.path.dirname(defaultfile) != '':
            open_dlg.SetDirectory(os.path.dirname(defaultfile))

    path = ''
    wc = ''
    if open_dlg.ShowModal() == wx.ID_OK:
        path = open_dlg.GetPath()
        wc = open_dlg.GetFilterIndex()
        open_dlg.Destroy()
    open_dlg.Destroy()
    if return_filterindex:
        return path, wc
    else:
        return path


def writedir(parent=None,  message='Directory to write'):

    dlg = wx.DirDialog(parent, message=message)
    path = ''
    wc = ''
    if dlg.ShowModal() == wx.ID_OK:
        path = dlg.GetPath()
        dlg.Destroy()
        return path
    else:
        dlg.Destroy()
        return None


def readdir(parent=None,  message='Directory to read'):

    dlg = wx.DirDialog(parent, message=message)
    path = ''
    wc = ''
    if dlg.ShowModal() == wx.ID_OK:
        path = dlg.GetPath()
        dlg.Destroy()
        return path
    else:
        dlg.Destroy()
        return None


def textentry(parent=None, message='', title='', def_string='', center=False,
              center_on_screen=False):
    dlg = TextEntryDialog(parent,
                          message, caption=title, value=def_string)
    if center:
        dlg.Centre()
    if center_on_screen:
        dlg.CentreOnScreen()
    if dlg.ShowModal() == wx.ID_OK:
        new_name = str(dlg.GetValue())
        dlg.Destroy()
        return True, new_name

    dlg.Destroy()
    return False, ''


def textselect(parent=None, message='', title='', def_string='',
               center=False, choices=[''],
               center_on_screen=False,
               endmodal_on_lastvalue=False):
    s = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
         "choices": choices}

    if def_string == '':
        def_string = choices[0]
    if message != '':
        ll = [[None, message, 2, None],
              ['Selection', def_string,  104,  s], ]
    else:
        ll = [['Selection', def_string,  104,  s], ]
    from ifigure.utils.edit_list import EditListDialog

    if endmodal_on_lastvalue:
        endmodal_lastvalue = endmodal_on_lastvalue
    else:
        endmodal_lastvalue = None
    dlg = EditListDialog(parent, wx.ID_ANY, title, ll,
                         endmodal_value=endmodal_lastvalue)
    if center:
        dlg.Centre()
    if center_on_screen:
        dlg.CentreOnScreen()
    val = dlg.ShowModal()
    value = dlg.GetValue()
    #print(val, wx.ID_OK)
    if val == wx.ID_OK:
        dlg.Destroy()
        return True, value[-1]
    dlg.Destroy()
    return False, value[-1]


def message(parent=None, message='', title='', style=0,
            icon=wx.ICON_EXCLAMATION,
            center_on_screen=False,
            center_on_parent=False,
            labels=None):
    if style == 0:
        style0 = wx.OK
    if style == 1:
        style0 = wx.CANCEL
    if style == 2:
        style0 = wx.OK | wx.CANCEL
    if style == 3:
        style0 = wx.OK | wx.NO
    if style == 4:
        style0 = wx.YES | wx.NO
    if style == 5:
        style0 = wx.YES | wx.NO | wx.CANCEL

#    from wx.lib.agw.genericmessagedialog import GenericMessageDialog as dia
    from wx import MessageDialog as dia
    style0 = style0 | icon

    dlg = dia(parent,
              message,
              title,
              style0)
    if labels is not None:
        dlg.SetOKCancelLabels(labels[0], labels[1])
    if center_on_screen:
        dlg.CentreOnScreen()
    if center_on_parent:
        dlg.CentreOnParent()

    ret = dlg.ShowModal()
    if ret == wx.ID_OK:
        ret = 'ok'
    elif ret == wx.ID_CANCEL:
        if style != 3:
            ret = 'cancel'
        else:
            ret = 'no'
    elif ret == wx.ID_YES:
        ret = 'yes'
    elif ret == wx.ID_NO:
        ret = 'no'
    dlg.Destroy()
    return ret


def showtraceback(parent=None, txt='', title="Error", traceback='None\n',
                  center_on_screen=True):
    from ifigure.widgets.dlg_message_scroll import DlgMessageScroll

    dlg = DlgMessageScroll(parent, wx.ID_ANY, title)
    if center_on_screen:
        dlg.CentreOnScreen()
    dlg.update(txt + '\n'+traceback)
    ret = dlg.ShowModal()

    dlg.Destroy()


def progressbar(parent, message, title, count):
    dlg = wx.ProgressDialog(title, message, count, parent)

    def close_dlg(evt, dlg=dlg):
        dlg.Destroy()
    dlg.Bind(wx.EVT_CLOSE, close_dlg)
    return dlg



import sys
import time
import tempfile
import traceback
from ifigure.utils.pickled_pipe import PickledPipe
from .utils import parse_server_string


class JobRunner(object):
    def __init__(self):
        self.use_mdsconnect = False
        self.connection_str = ''
        self.connection = None
        self.g = {}
        self._shot = -1

    def run(self, jobset, skip_init):
        import MDSplus
        from MDSplus import Connection

        self.error = []
        jobs, names = jobset
        self.r = {}
        self.r['error message'] = ['']
        try:
            for k, job in enumerate(jobs):
                for j in job:
                    if j.command == 'connection_mode':
                        if self.connection_str != j.params[0]:
                            self.connection_str = j.params[0]
                            server, port, tree = parse_server_string(
                                self.connection_str)
                            self.use_mdsconnect = (False if server.upper() ==
                                                   'DIRECT' else True)
                            if self.use_mdsconnect:
                                if self.connection is not None:
                                    del self.connection
                                try:
                                    if port != '':
                                        server = server + ':' + port
                                    self.connection = Connection(server)
                                    res = 'ok'
                                except:
                                    res = ('connection error' +
                                           traceback.format_exc())
                            else:
                                if self.connection is not None:
                                    del self.connection
                                    self.connection = None
                                res = 'ok'
                        else:
                            res = 'ok'
                    else:
                        if j.command == 'init':
                            if skip_init:
                                continue
                            else:
                                j.command = 'value'
                        if self.use_mdsconnect:
                            res = self._run_job_mdsconnect(j)
                            if len(self.error) != 0:
                                self.r['error message'] = self.error
                                self.r[names[k]] = res
                                return self.r
                        else:
                            res = self._run_job(j)
                            if len(self.error) != 0:
                                self.r['error message'] = self.error
                                self.r[names[k]] = res
                                return self.r
                self.r[names[k]] = res
        except:
            self.error = ['run error', traceback.format_exc()]
        self.r['error message'] = self.error
        return self.r

    def _run_script_txt(self, job):
        expr = job.params[0]
        try:
            code = compile(expr, '<string>', 'exec')
            g = {}
            l = {}
            exec(code, self.g, self.r)
            return 'ok'
        except:
            self.error = ['Scrip Error', expr, sys.exc_info()[0]]
        return None

    def _run_job_mdsconnect(self, job):
        com = job.command
        if com == 'open':
            tree = job.params[0]
            shot = job.params[1]
            try:
                self.connection.openTree(tree, int(shot))
                return 'ok'
            except:
                self.error = ['run error', traceback.format_exc()]
                return None
            self._shot = shot
        elif com == 'defnode':
            node = job.params[0]
            try:
                if node.strip() != '':
                    self.connection.setDefault(node)
                return 'ok'
            except:
                self.error = ['run error', traceback.format_exc()]
                return None
        elif com == 'value':
            try:
                expr = job.params[0]
                r = self.connection.get(expr).data()
                return r
            except:
                self.error = ['run error', expr, traceback.format_exc()]
                return None
        elif com == 'script_txt':
            return self._run_script_txt(job)

    def _run_job(self, job):
        import MDSplus
        from MDSplus import Connection

        com = job.command

        if com == 'open':
            tree = job.params[0]
            shot = job.params[1]
            try:
                self.t = MDSplus.Tree(tree, int(shot))
                return 'ok'
            except:
                self.error = ['run error', traceback.format_exc()]
                return None
            self._shot = shot
        elif com == 'defnode':
            node = job.params[0]
            try:
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                return 'ok'
            except:
                self.error = ['run error', traceback.format_exc()]
                return None
        elif com == 'value':
            try:
                expr = job.params[0]
                r = MDSplus.Data.compile(expr).evaluate().data()
                return r
            except:
                self.error = ['run error', expr, traceback.format_exc()]
                return None
        elif com == 'script_txt':
            return self._run_script_txt(job)


if __name__ == '__main__':
    ch = PickledPipe(sys.stdin, sys.stdout)
    runner = JobRunner()
    while True:
        time.sleep(0.01)
        try:
            jobset = ch.recv(nowait=False)
        except EOFError:
            break
        if jobset != -1:
            tmp = tempfile.TemporaryFile()
            sys.stdout = tmp
            try:
                r = runner.run(jobset)
            except:
                r = {'error message': ['runner.run failed']}
#               r = {'error message':['message was received']}
                r['worker output'] = ''
                ch.send(r)
            finally:
                tmp.flush()
                tmp.seek(0)
                r['worker output'] = tmp.read()
                tmp.close()

                jobs, names = jobset
                def_names = ['worker output',
                             'error message',
                             'x', 'y', 'z',
                             'xerr', 'yerr', 'zerr']
                for n in names:
                    if not n in def_names:
                        def_names.append(n)
                r2 = {}
#               r = {n:r[n] for n in def_names if r.has_key(n)}
                for n in def_names:
                    if n in r:
                        r2[n] = r[n]

                ch.send(r2)
        else:
            break



#!/usr/bin/python

from __future__ import print_function
import socket
import threading
import os
import signal
import sys
import time
from six.moves import socketserver
import paramiko
import binascii
import ifigure.utils.pickle_wrapper as pickle
from ifigure.utils.daemon import Daemon
import MDSplus


class ThreadingTCPRequestHandler(socketserver.BaseRequestHandler):
    #
    def __init__(self, *args, **kargs):
        socketserver.BaseRequestHandler.__init__(self, *args, **kargs)
        name = '/tmp/mdspluse_gateway_thread'+str(os.getpid()) + '.log'
#        self.fid = open(name, 'w')

    def handle(self):
        message = self.request.recv(1024)

        data = pickle.loads(binascii.a2b_hex(message))
        #cur_thread = threading.current_thread()
        print(("Request ", data))

        r = {}
        for name, commands in data:
            for command in commands:
                com = command[0]
                param = command[1:]
                r[name] = self._handle(com, param)

        print(('return variables', list(r.keys())))

        sr = pickle.dumps(r)
        response = binascii.b2a_hex(sr)
        print(("sending data (length)", len(response)))
        self.request.sendall(response+'\n')
#        self.fid.flush()

    def _handle(self, com, param):
        response = ''

        print((com, param))
        if com == 'c':  # connection request
            tree, shot, node = param.split(',')
            try:
                print(('opening tree', tree, shot))
                self.t = MDSplus.Tree(tree, int(shot))
                print(('opening tree', self.t))
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                response = 'ok'
            except Exception:
                print(('Error', str(sys.exc_info()[
                      1])+','+str(sys.exc_info()[2])))
                response = ''
        if com == 'd':
            node = param
            print(('setting def node', node))
            try:
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                response = 'ok'
            except Exception:
                print(('Error', str(sys.exc_info()[
                      1])+','+str(sys.exc_info()[2])))
                response = ''

        if com == 'f':  # connection request
            a = param.split(',')
            tree = a[0]
            shot = a[1]
            node = a[2]
            expr = ','.join(a[3:])
            try:
                self.t = MDSplus.Tree(tree, int(shot))
                if node.strip() != '':
                    tn = self.t.getNode(node)
                    self.t.setDefault(tn)
                response = MDSplus.Data.compile(expr).evaluate().data()
#              print "sending data (length)", len(response)
            except Exception:
                print(('Error', str(sys.exc_info()[
                      1])+','+str(sys.exc_info()[2])))
                response = ''

        if com == 'v':  # mdsvalue
            print(('mdsvalue', param))
            try:
                response = MDSplus.Data.compile(param).evaluate().data()
            except Exception:
                print(('Error', str(sys.exc_info()[
                      1])+','+str(sys.exc_info()[2])))
                response = ''

#           print "sending data (length)", len(response), 'original :', type(sr)

        return response

class ThreadingTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class MDSDaemon(Daemon):
    def __init__(self, HOST, PORT, pidfile, stdout='/dev/null',
                 stderr='/dev/null'):
        self.host = HOST
        self.port = PORT
        Daemon.__init__(self, pidfile, stdout=stdout, stderr=stderr)

    def run(self):
        self.server = ThreadingTCPServer((self.host, self.port),
                                         ThreadingTCPRequestHandler)

        print(('Starting MDSGateway', self.host, self.port))
        print(sys.stdout)
        self.ip, self.port = self.server.server_address
        sys.stdout.flush()
        sys.stderr.flush()

        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        # Exit the server thread when the main thread terminates
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        while True:
            # this loop is to keep thread on
            time.sleep(0.1)
            sys.stdout.flush()
            sys.stderr.flush()

    def signal_handler(self, signal, frame):
        print(('Received singal', signal))
        self.server.shutdown()
        sys.exit(0)

    def stop(self):
        print('Stopping server')
        Daemon.stop(self)


if __name__ == "__main__":
    HOST, PORT = socket.gethostname(), 10002
    daemon = MDSDaemon(HOST, PORT, '/tmp/mdsgateway.pid',
                       stdout='/tmp/mdsgateway.log',
                       stderr='/tmp/mdsgateway.log')

    if len(sys.argv) == 2:
        if 'start' == sys.argv[1]:
            daemon.start()
        elif 'stop' == sys.argv[1]:
            daemon.stop()
        elif 'restart' == sys.argv[1]:
            daemon.restart()
        else:
            print("Unknown command")
            sys.exit(2)
            sys.exit(0)
    else:
        print("usage: %s start|stop|restart" % sys.argv[0])
        sys.exit(2)



from ifigure.mto.fig_obj import FigObj
from ifigure.mto.generic_points import GenericPoint, GenericPointsHolder


class FigObjGPHolder(FigObj, GenericPointsHolder):
    def destroy(self, clean_owndir=True):
        if self._cb_added:
            fig_page = self.get_figpage()
            fig_page.rm_resize_cb(self)
        GenericPointsHolder.destroy(self)
        FigObj.destroy(self, clean_owndir=clean_owndir)

    def check_loaded_gp_data(self):
        figpage = self.get_figpage()
        figaxes = self.get_figaxes()

        for k, gp in enumerate(self._gp_points):
            gp.set_figpage(figpage)

        if hasattr(self, '_loaded_gp_data'):
            for k, gp in enumerate(self._gp_points):
                gp.set_figpage(figpage)
                gp.dict_to_gp(self._loaded_gp_data[k], self)
            del self._loaded_gp_data

        for k, gp in enumerate(self._gp_points):
            if figaxes is not None:
                gp.set_figaxes(figaxes)

    def set_parent(self, parent):
        FigObj.set_parent(self, parent)
        GenericPointsHolder.set_parent(self, parent)

    def save_data(self, fid=None):
        FigObj.save_data(self, fid)
        GenericPointsHolder.save_data(self, fid)

    def load_data(self, fid=None):
        FigObj.load_data(self, fid)
        GenericPointsHolder.load_data(self, fid)

    def save_data2(self, data):
        data = FigObj.save_data2(self, data)
        data = GenericPointsHolder.save_data2(self, data)
        return data

    def load_data2(self, data):
        FigObj.load_data2(self, data)
        GenericPointsHolder.load_data2(self, data)



# this file is incomplete

mode = 'emacs'
code_table = {7801: 'C-N',
              8001: 'C-P',
              6801: 'C-X',
              7001: 'C-P',
              6601: 'C-B',
              8901: 'C-Y',
              8701: 'C-W',
              8702: 'A-W',
              8302: 'C-S',
              9001: 'C-Z',
              6901: 'C-E',
              7501: 'C-K',
              6501: 'C-A', }


def controlDown(key_evt):
    if hasattr(key_evt, 'RawControlDown'):
        controlDown = key_evt.RawControlDown()
    else:
        controlDown = key_evt.ControlDown()
    return controlDown


def interpret_keycombination(key_evt):

    if hasattr(key_evt, 'RawControlDown'):
        mod = (key_evt.RawControlDown() +
               key_evt.AltDown()*2 +
               key_evt.ShiftDown()*4 +
               key_evt.MetaDown()*8)
    else:
        mod = (key_evt.ControlDown() +
               key_evt.AltDown()*2 +
               key_evt.ShiftDown()*4 +
               key_evtprev.MetaDown()*8)

    code = key_evt.GetKeyCode()

    if code*100+mod in code_table:
        return code_table[code*100+mod]
    return None



from __future__ import print_function

'''
    client.py


    it provides server/client operation of piScope

    A user can launch piScope from any python console,
    and send/recieve various command/data.

    useage:

        from ifigure.client import *

        launch()        # this launch piScope as a separate process
        plot(range(30)) # send range(30) to piscope and make a plot

        ## here a user export plot data to piScope shell, and
        ## create xdata and ydata

        print xdata     # in server/client mode, the data is automatically 
                        # sent

        ## a user can also get any pickable variables in piScope shell
        get('varname')

'''

import socket
import subprocess
import sys
import shlex
import ifigure.utils.pickle_wrapper as cPickle
import binascii
import threading
import os
from six.moves import socketserver
from ifigure.utils.cbook import pick_unused_port


class ReceiverReqHandler(socketserver.BaseRequestHandler):
    def handle(self):
        rfile = self.request.makefile('r')
        response = rfile.readline().strip()
        rfile.close()
        data = cPickle.loads(binascii.a2b_hex(response))
        if data['type'] == 'data':
            import __main__
            text = '\n'
            for key in data['data'][0]:
                if key in dir(__main__):
                    text = text + key + ' is updated\n'
                else:
                    text = text + key + ' is created\n'
                setattr(__main__, key, data['data'][0][key])
            print(text)
        elif data['type'] == 'msg':
            print((data['data']))


class Receiver(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class Client(object):
    port = 0
    host = 'localhost'
    process = None
    receiver = None

    def launch(self, host='localhost', exe=None):
        if host == 'localhost':
            from ifigure.utils.get_ifigure_dir import bin_dir
            # command = os.path.join(bin_dir(), 'piscope.sh') + ' -s'
            if exe is None:
                exe = sys.executable
            # command = command + ' -e '+ exe + ' &'
            # import piscope
            # command = sys.executable + ' ' + piscope.__file__ + ' -s -d'
            command = 'piscope  -s -d'
            # print(command)
            if os.altsep is not None:
                command = command.replace(os.sep, os.altsep)
            p = subprocess.Popen(shlex.split(command),  # shell = True,
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True)
            lhost = 'localhost'
        else:
            pass
        line = ''
        while line[0:5] != 'start':
            line = p.stdout.readline()
            print(line)
        arr = line.split(':')
        Client.host = arr[1].rstrip("\r\n").strip()
        Client.port = int(arr[2].rstrip("\r\n").strip())
        Client.process = p
        if Client.receiver is None:
            port = pick_unused_port()

            sys.stdout.flush()
            Client.receiver = Receiver((lhost, port), ReceiverReqHandler)

            server_thread = threading.Thread(
                target=Client.receiver.serve_forever)
            server_thread.daemon = True
            server_thread.start()
        ip, port = Client.receiver.server_address
        print(('receiver :', ip, ':', port))
        message = cPickle.dumps(('r', ip, port))
        self.send(message, noresponse=True)

    def shutdown(self):
        if Client.process is not None:
            Client.process.kill()
            Client.process = None
            Client.host = 'localhost'
            Client.port = 0
            Client.receiver.shutdown()
            Client.receiver = None

    def set_connection(self, host, port):
        Client.port = port
        Client.host = host

    def send(self, message, noresponse=False):
        host = Client.host
        port = Client.port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))
        response = None
        try:
            hexmessage = binascii.b2a_hex(message)
            sock.sendall(hexmessage + b'\n')
            if not noresponse:
                rfile = sock.makefile('r')
                response = rfile.readline().strip()
                rfile.close()
#             response = sock.recv(1024)
#             print len(response)
                response = cPickle.loads(binascii.a2b_hex(response))
        finally:
            sock.close()
        return response


def launch(exe=None):
    server('launch', exe=exe)


def shutdown():
    server('shutdown')


def server(param, host='localhost', port=None, exe=None):
    '''
    launch/connect piscope 
    server('coonect', host, port)
    server('launch')
    server('shutdown')
    '''
    c = Client()
    if param == 'launch':
        c.launch(exe=exe)
    if param == 'connect':
        c.set_connection(host, port)
    if param == 'shutdown':
        if c.host is None:
            return
        if c.port == 0:
            return

        message = cPickle.dumps(('f', 'quit', tuple(), dict()))
        c = Client()
        c.send(message, noresponse=True)
        c.shutdown()
    print(('host: ', c.host, ', port: ', c.port))


def check_connection():
    c = Client()
    message = cPickle.dumps(('c',))
    print((c.send(message)))


def make_testplot():
    message = cPickle.dumps(('t', 'plot(range(10))'))
    c = Client()
    c.send(message)


def _get_random_name():
    # get random name
    from time import gmtime, strftime
    import hashlib
    import os
    import ifigure_config

    check = True
    while(check):
        strtime = strftime("%Y-%m-%d %H:%M:%S", gmtime())
        m = hashlib.md5()
        m.update(strtime)
        txt = m.hexdigest()
        fpath = os.path.join(ifigure_config.rcdir,
                             '###param_'+txt)
        check = os.path.exists(fpath)
    return fpath


def _save_parameter_file(*args, **kargs):
    try:
        return sr
    except IOError as error:
        return False


def _send_message(command, *args, **kargs):
    try:
        message = cPickle.dumps(('f', command, args, kargs))
    except error:
        print('failed to save parameter file')
        return
    c = Client()
    return c.send(message)


def _send_message_g(command, *args, **kargs):
    try:
        message = cPickle.dumps(('g', command, args, kargs))
    except error:
        print('failed to save parameter file')
        return
    c = Client()
    return c.send(message)


def cla(*args, **kargs):
    _send_message('cla', *args, **kargs)


def cls(*args, **kargs):
    _send_message('cls', *args, **kargs)


def clf(*args, **kargs):
    _send_message('clf', *args, **kargs)


def nsec(*args, **kargs):
    _send_message('nsec', *args, **kargs)


def nsection(*args, **kargs):
    _send_message('nsection', *args, **kargs)


def isec(*args, **kargs):
    _send_message('isec', *args, **kargs)


def isection(*args, **kargs):
    _send_message('isection', *args, **kargs)


def scope(*args, **kargs):
    _send_message('scope', *args, **kargs)


def figure(*args, **kargs):
    _send_message('figure', *args, **kargs)


def addpage(*args, **kargs):
    _send_message('addpage', *args, **kargs)


def delpage(*args, **kargs):
    _send_message('delpage', *args, **kargs)


def hold(*args, **kargs):
    _send_message('hold', *args, **kargs)


def update(*args, **kargs):
    _send_message('update', *args, **kargs)


def title(*args, **kargs):
    _send_message('title', *args, **kargs)


def xlabel(*args, **kargs):
    _send_message('xlabel', *args, **kargs)


def ylabel(*args, **kargs):
    _send_message('ylabel', *args, **kargs)


def xlim(*args, **kargs):
    _send_message('xlim', *args, **kargs)


def ylim(*args, **kargs):
    _send_message('ylim', *args, **kargs)


def subplot(*args, **kargs):
    _send_message('subplot', *args, **kargs)


def plot(*args, **kargs):
    _send_message('plot', *args, **kargs)


def errorbar(*args, **kargs):
    _send_message('errorbar', *args, **kargs)


def contour(*args, **kargs):
    _send_message('contour', *args, **kargs)


def image(*args, **kargs):
    _send_message('image', *args, **kargs)


def ispline(*args, **kargs):
    _send_message('ispline', *args, **kargs)


def axline(*args, **kargs):
    _send_message('axline', *args, **kargs)


def axspan(*args, **kargs):
    _send_message('axspan', *args, **kargs)


def text(*args, **kargs):
    _send_message('text', *args, **kargs)


def legend(*args, **kargs):
    _send_message('legend', *args, **kargs)


def get(*args, **kargs):
    return _send_message_g('get_shellvar', *args, **kargs)


def put(*args, **kargs):
    return _send_message_g('pet_shellvar', *args, **kargs)



#
#   this is for the memo of the routines to be
#   implemented for widgets used in propertyeditor
#
#


class base_widget(object):
    def __init__(self):
        self.property = ''

    def SetEditorValue(self,  artist=None):
        # this shouuld define how to update
        # gui based on the data in ifigure_canvas
        pass

    def SetCanvasValue(self,  artist=None):
        pass
        # this shouuld define how to update
        # ifigure_canvas based on GUI



from __future__ import print_function
from ifigure.utils.images2gif import writeGif
#from PIL import Image
import os
from ifigure.utils.cbook import image_to_pil
import six


def save_animation(func, params, canvas, filename='animation.gif', duration=0.2):
    images = []
    for p in params:
        func(p)
        image = canvas.canvas.bitmap.ConvertToImage()
        images.append(image_to_pil(image))

    # we use PIL's implementation in PY3
    images[0].save(filename,
                   save_all=True,
                   append_images=images[1:],
                   optimize=False,
                   duration=duration*1000,  # this parameter is ms
                   loop=0)



#  Name   :fig_text
#
#          text object.
#          surpport both text and figtext
#          command.
#
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- S. Shiraiwa
# *******************************************
import matplotlib.lines
import matplotlib.transforms as mpltransforms
import numpy as np
import weakref
import os

from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_axes import FigAxes
from ifigure.mto.fig_page import FigPage
from ifigure.widgets.canvas.file_structure import *
import ifigure
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
import ifigure.utils.geom as geom

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod

from ifigure.mto.generic_points import GenericPoint, GenericPointsHolder
from ifigure.mto.figobj_gpholder import FigObjGPHolder

from ifigure.utils.args_parser import ArgsParser
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigText')
num_gp = 1


class FigText(FigObjGPHolder):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []  # for debug....
            obj._drag_backup = None
            obj._drag_mode = 1  # 1 transpose 2 expand 3 rotate
            obj._drag_start = None
            return obj

        if 'src' in kywds:
            obj = FigObjGPHolder.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            obj.setvar('clip', True)  # backword compatiblity should come here
            return obj

        p = ArgsParser()
        p.add_opt('x', 0.5, ['number', 'dynamic'])
        p.add_opt('y', 0.5, ['number', 'dynamic'])
        p.add_opt('z', 0.0, ['number', 'dynamic'])
        p.add_var('s', ['str', 'dynamic'])
        p.set_pair("x", "y")
        p.add_key('draggable', True)
        p.add_key('bbox', dict())
        p.add_key("fancybox", False)
        p.add_key("fbcolor", 'red')
        p.add_key("fbecolor", 'red')
        p.add_key("fbstyle", 'square')
        p.add_key("fbpad", 0.3)
        p.add_key("fbalpha", 1)
        p.add_key("fontfamily", 'sans-serif')
        p.add_key("fontstyle",  'normal')
        p.add_key("fontweight", 'roman')
        p.add_key("multialignment", "left")
        p.add_key("clip", True)
        p.add_key("trans", ["figure", "figure"]*num_gp)
        p.add_key("transaxes", ["default"]*num_gp)

        v, kywds, d, flag = p.process(*args, **kywds)

        if not flag:
            raise ValueError('Failed when processing argument')

        if isinstance(v['trans'], str):
            v['trans'] = [v['trans'], ]*num_gp*2

        obj = FigObjGPHolder.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)

        return obj

    def __init__(self, *args,  **kywds):
        self._data_extent = None
        self._cb_added = False
        self._2d_text = True
        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        GenericPointsHolder.__init__(self, num=num_gp)
        super(FigText, self).__init__(*args, **kywds)

    @classmethod
    def isFigText(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def allow_outside(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'text'

    @classmethod
    def property_in_file(self):
        return ["size", ] + \
            ["text", "color", "rotation", "alpha", "zorder"]

    @classmethod
    def property_in_palette(self):
        return ["text", "box"], [["text", "size",
                                  "color", "fontfamily", "fontweight", "fontstyle",
                                  "rotation", "alpha", "multialignment"],
                                 ["fancybox", "fbcolor", "fbecolor", "fbstyle", "fbpad",
                                  "fbalpha"]]

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ["use_var", "fancybox", "fbcolor", "fbecolor",
                "fbstyle", "fbpad", "fbalpha", "zorder",
                "fontstyle", "fontweight", "fontfamily",
                "multialignment", "clip"]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'a.png')
        return [idx1]

    def set_parent(self, parent):
        super(FigText, self).set_parent(parent)
#        GenericPointsHolder.set_gp_figpage(self)
        if isinstance(parent, FigPage):
            self._isFigPage = True
        else:
            self._isFigPage = False

    def isDraggable(self):
        return self._var["draggable"]

    def args2var(self):
        names0 = self.attr_in_file()
#        names  = ["x","y", "z", "s"]
#        use_np = [False]*4
        names = ["s"]
        use_np = [False]
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)

        self.set_gp_by_vars()  # use trans, transaxes in vars

#        x = values[0];y = values[1]
#        if x is None: return False
#        if y is None: return False
#        if (x.size != 1) or (y.size != 1):
#            self.setp("x", None)
#            self.setp("y", None)
#            return False
        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change

        if self.isempty() is False:
            return
#              self.del_artist(delall=True)

        lp = self.getp("loaded_property")

        # this will take care of use_var
        # and intialize gp points from vars if use_var
        x, y, z, s = self._eval_s()
        kywds = self._var["kywds"]

        if self.getp("fancybox"):
            style = self.getp("fbstyle")
            pad = self.getp("fbpad")
            t = style+',pad='+str(pad)
            bbox_props = dict(boxstyle=t,
                              facecolor=self.getp("fbcolor"),
                              edgecolor=self.getp("fbecolor"),
                              alpha=self.getp("fbalpha"))
            self.setp("bbox", bbox_props)
            kywds['bbox'] = bbox_props
        else:
            if 'bbox' in kywds:
                del kywds['bbox']
        if lp is not None:
            s = ''
#              x, y, s = self.eval("x", "y", "s")
#              if (x is not None  and
#                  y is not None):
#                 self.set_gp_point(0, x, y)
#           else:
#              s = ''
#              if len(lp) > 1:
#                 self.get_gp(0).dict_to_gp(lp[-1][0], self)
#              else:
#                 self.set_gp_figpage()

        self.make_newartist(s, **kywds)

        if not self._cb_added:
            fig_page = self.get_figpage()
            fig_page.add_resize_cb(self)
            self._cb_added = True

    def onResize(self, evt):
        self.refresh_artist()

    def do_update_artist(self):
        #        print 'text::do_update_artists'
        self.refresh_artist()

    def make_newartist(self, s='', **kywds):
        self.check_loaded_gp_data()

        container = self.get_container()
        xd, yd = self.get_device_point(0)
        # print xd, yd, self.get_gp(0).x, self.get_gp(0).y

        kywds['family'] = str(self.getp('fontfamily'))
        kywds['style'] = str(self.getp('fontstyle'))
        kywds['weight'] = str(self.getp('fontweight'))

        # multialingment and horizontal alignment
        # should be the same, somehow...(2014 01 07)

        kywds['multialignment'] = str(self.getp('multialignment'))
        kywds['ha'] = str(self.getp('multialignment'))

        if self._2d_text:
            self._artists = [container.text(xd, yd, s, **kywds)]
            self._artists[0].set_transform(mpltransforms.IdentityTransform())
        else:
            x, y, z, s = self._eval_s()
            self._artists = [container.text(x, y, z, s, **kywds)]

        lp = self.getp("loaded_property")
        if lp is not None:
            self.set_artist_property(self._artists[0], lp[0])
#        self._artists[0].set_size(self.getp('size'))
        self._artists[0].figobj = self
        self._artists[0].figobj_hl = []
        self._artists[0].set_zorder(self.getp('zorder'))
        if self.get_figaxes() is not None and self.getp('clip'):
            bbox = mpltransforms.Bbox.from_extents(
                container.get_window_extent().extents)
            try:
                self._artists[0].set_clip_box(bbox)
                self._artists[0].set_clip_on(True)
                self._artists[0]._bbox_patch.set_clip_box(bbox)
                self._artists[0]._bbox_patch.set_clip_on(True)
            except:
                pass

        self.delp("loaded_property")

        return self._artists[0]

    def refresh_artist(self):
        if len(self._artists) != 1:
            return
        a1 = self._artists[0]
        z = a1.zorder
        hl = len(self._artists[0].figobj_hl) != 0
        self.del_artist(delall=True)
        self.generate_artist()
        a2 = self._artists[0]
        #a2 = self.make_newartist()
        a2.set_zorder(z)
        self.highlight_artist(hl, [a2])
        if a1 != a2:
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def refresh_artist_data(self):
        if self.isempty() and not self._suppress:
            self.generate_artist()
            return
        x, y, z, s = self._eval_s()  # this handles "use_var"
        self._artists[0].set_text(s)
        if x is not None and y is not None:
            #           self.set_gp_point(0, x, y)
            xd, yd = self.get_device_point(0)
            self._artists[0].set_x(xd)
            self._artists[0].set_y(yd)

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        # save_data2->load_data2 will set "loaded_property"
        self.store_loaded_property()

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            is_figtext = self.get_figaxes() is None
            for a in artistlist:
                #             a.set_picker(None)
                #            this does not work for figtext
                #             a.remove()
                if is_figtext:
                    container.texts.remove(a)
                else:
                    a.remove()

        super(FigText, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist

        page = self.get_figpage()
        figure = page._artists[0]
        container = figure
        #container = self.get_container()

        if container is None:
            return

        if val == True:
            for a in alist:
                try:
                    box = a.get_window_extent().get_points()
                    x = [box[0][0], box[0][0], box[1][0], box[1][0], box[0][0]]
                    y = [box[0][1], box[1][1], box[1][1], box[0][1], box[0][1]]
                except:
                    # if get_window_extent fails, it drow somthing
                    xd, yd = self.get_device_point(0)
                    x = [xd, xd+3, xd+3, xd, xd]
                    y = [yd, yd, yd+3, yd+3, yd]

                hl = matplotlib.lines.Line2D(x, y, marker='s',
                                             color='k', linestyle='None',
                                             markerfacecolor='None',
                                             markeredgewidth=0.5,
                                             figure=container)
                container.lines.extend([hl])
                self._objs.append(weakref.ref(hl))
                a.figobj_hl.append(hl)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    container.lines.remove(hl)
                a.figobj_hl = []
#
#   def hit_test
#

    def get_artist_extent(self, a):
        try:
            box = a.get_window_extent().get_points()
            return [box[0][0], box[1][0], box[0][1], box[1][1]]
        except:
            # if get_window_extent fails, it drow somthing
            xd, yd = self.get_device_point(0)
            return [xd, yd, xd+3, yd+3]

    def dragstart_a(self, a, evt):
        redraw = super(FigText, self).dragstart_a(a, evt)
        self._drag_backup = (a.get_position(), a.get_size())
        if self._picker_a_type == 'edge':
            self._picker_a_type = 'area'
        return redraw

    def drag_update_textpos(self, a):
        sx = min([self._st_extent[0], self._st_extent[1]])
        sy = min([self._st_extent[2], self._st_extent[3]])
        px = min([self._drag_rec[0], self._drag_rec[1]])
        py = min([self._drag_rec[2], self._drag_rec[3]])

        sa = np.array([sx, sy]).reshape(1, 2)
        sd = a.get_transform().inverted().transform(sa)
        pa = np.array([px, py]).reshape(1, 2)
        pd = a.get_transform().inverted().transform(pa)
        x = pd.flat[0] - sd.flat[0] + self._drag_backup[0][0]
        y = pd.flat[1] - sd.flat[1] + self._drag_backup[0][1]

        r = (float(self._drag_rec[1]-self._drag_rec[0]) /
             float(self._st_extent[1]-self._st_extent[0]))

        return x, y, self._drag_backup[1]*abs(r)

    def drag_a(self, a, evt, shift=None, scale=None):
        #        print  self._picker_a_type
        if self._picker_a_type == 'area':
            shift = evt.guiEvent.ShiftDown()
        else:
            shift = True
        redraw, scale = super(FigText, self).drag_a(a, evt,
                                                    shift=shift, scale=scale)

        return redraw, scale

#    def dragdone_a_clean(self, a):
#        a=self._artists[0]
#        a.set_x(self._drag_backup[0][0])
#        a.set_y(self._drag_backup[0][1])
#        a.set_size(self._drag_backup[1])
#        super(FigText, self).dragdone_a_clean(a)

    def dragdone_a(self, a, evt, sfift=None, scale=None):
        if self._picker_a_type == 'area':
            shift = evt.guiEvent_memory.ShiftDown()
        else:
            shift = True
        redraw, scale0 = super(FigText, self).dragdone_a(a, evt,
                                                         shift=shift, scale=scale)

        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()

#        hist = self.get_root_parent().app.history
        h = self.scale_artist(scale0, action=True)
        GlobalHistory().get_history(window).make_entry(h)
        return 0, scale0

    def scale_artist(self, scale, action=True):
        st_extent = self.get_artist_extent(self._artists[0])
        rec = geom.scale_rect(st_extent,
                              scale)
        a1 = self.move_gp_points(0, rec[0]-st_extent[0],
                                 rec[2]-st_extent[2], action=True)
        a = self._artists[0]
        s = a.get_size()*(scale[0]+scale[3])/2
        a2 = UndoRedoArtistProperty(a, 'size', s)
        return [a2, a1]

    def canvas_menu(self):
        def onClipTextByAxes(evt, figobj=self):
            action1 = UndoRedoFigobjProperty(figobj._artists[0],
                                             'clip', not self.getp('clip'))
            window = evt.GetEventObject().GetTopLevelParent()
            hist = GlobalHistory().get_history(window)
            hist.make_entry([action1], menu_name='clipping')
        if not self._2d_text:
            return []
        m = self.gp_canvas_menu(self.get_gp(0))
        if self.get_figaxes() is not None:
            if self.getp('clip'):
                txt = "^Clip by axes"
            else:
                txt = "*Clip by axes"
            m.extend([(txt, onClipTextByAxes, None), ])
        return m
#        return self.gp_holder_canvas_menu()

    def load_data(self, fid=None):
        FigObj.load_data(self, fid)
#        GenericPointsHolder.load_data(self, fid)

    def set_multialignment(self, value, a):

        ovalue = self.getp('multialignment')
        self.setp('multialignment', str(value))

        # changing multialinment also need to
        # change horizontal alignment of artists
        # this is to keep the position the same
        # after making two changes
        #  I am commenting out this adjustment since
        #  it may be better to change the postion and
        #  let a user to adjust it.
#        box=a.get_window_extent().get_points()
#        dx = abs(box[0][0] - box[1][0])
#        xd, yd = self.get_device_point(0)
#        d = {'left': 0, 'center':0.5, 'right':1}
#        shift = dx*(d[value] - d[ovalue])
#        self.set_device_point(0, xd+shift, yd)
        self.refresh_artist()

    def get_multialignment(self, a):
        return self.getp('multialignment')

    def save_data2(self, data=None):
        def check(obj, name):
            return obj.getp(name) is obj.getvar(name)
        if data is None:
            data = {}
        var = {
            #               'x':check(self,'x'),
            #               'y':check(self,'y'),
            #               'z':check(self,'z'),
            's': check(self, 's')}

#        if var["x"] is not True: var["xdata"] = self.getp("x")
#        if var["y"] is not True: var["ydata"] = self.getp("y")
#        if var["z"] is not True: var["zdata"] = self.getp("z")
        if var["s"] is not True:
            var["sdata"] = self.getp("s")

        data['FigText'] = (1, var)
        data = super(FigText, self).save_data2(data)
        return data

    def load_data2(self, data):
        if not self.hasvar('fontfamily'):
            self.setvar("fontfamily", 'sans-serif')
            self.setvar("fontstyle",  'normal')
            self.setvar("fontweight", 'roman')
            self.setp("fontfamily", 'sans-serif')
            self.setp("fontstyle",  'normal')
            self.setp("fontweight", 'roman')
        if not self.hasvar('multialignment'):
            self.setvar("multialignment", 'left')
            self.setp("multialignment", 'left')

        d = data['FigText']
        super(FigText, self).load_data2(data)
        var = d[1]
#        names = ["x", "y", "z", "s"]
        names = ["s"]

        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_s(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            x, y, z = self.eval("x", "y", "z")
            if (x is not None and
                    y is not None):
                self.set_gp_point(0, x, y)
            if not success:
                return None, None, None, None
        if self.get_figaxes() is not None:
            '''
            in case of 3d, it does not use generic_point,
            since the position of artist changes depending
            on view. This means that it can not be dragged
            by mouse...
            perhaps, I need to add three d object dragging
            in a generic way....
            '''
            if self.get_figaxes().get_3d():
                x, y, z = self.eval("x", "y", "z")
                self._2d_text = False
                return x, y, z, self.getp('s')

        x, y = self.get_gp_point(0)
        return [x, y, None, self.getp('s')]
#        return  self.getp(("x", "y", "z", "s"))

    def convert_to_tex_style_text(self, mode=True):
        from ifigure.utils.cbook import tex_escape_equation
        if len(self._artists) == 0:
            return
        if mode:
            s = self._artists[0].get_text()
            self.setp('s', s)
            self._artists[0].set_text(tex_escape_equation(str(s)))
        else:
            s = self.getp('s')
            self._artists[0].set_text(s)



from ifigure.utils.edit_list import DialogEditListTab
import ifigure.widgets.dialog as dialog
import wx
from ifigure.widgets.book_viewer import FrameWithWindowList
from ifigure.widgets.miniframe_with_windowlist import DialogWithWindowList
from ifigure.utils.edit_list import EditListPanel, EDITLIST_CHANGED
from ifigure.widgets.script_editor import Notebook
from ifigure.utils.wx3to4 import GridSizer

pkeys = ['experiment', 'default_node', 'x', 'y', 'z',
         'xerr', 'yerr', 'title', '_flag']


def make_gs_arr(gs, s):
    return {key: gs[key][s] for key in gs}
# class DlgMdsSetting(FrameWithWindowList):


class DlgMdsSetting(DialogWithWindowList):
    def __init__(self, parent, setting,  gwparam, mds_server, thread_main, cb=None):
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.CLOSE_BOX
        super(DlgMdsSetting, self).__init__(parent, wx.ID_ANY, style=style)
#      super(DlgMdsSetting, self).__init__(parent, wx.ID_ANY)
        self.cb = cb
        gs = gwparam.getvar('globalsets')
        gv = gwparam.getvar()
        tabs = ['Set1', 'Set2', 'Set3', 'Set4', 'config', 'more']

        list1 = [[None, make_gs_arr(gs, 0), 118, None]]
        list2 = [[None, make_gs_arr(gs, 1), 118, None]]
        list3 = [[None, make_gs_arr(gs, 2), 118, None]]
        list4 = [[None, make_gs_arr(gs, 3), 118, None]]

        txt = "on" if parent.mpanalysis else "off"
        txt2 = "on" if gv['use_shot_global'] else "off"
        txt3 = "on" if parent.book._use_global_tdi else "off"
        txt4 = parent.book.getvar('global_tdi')
        txt8 = parent.book.getvar('global_tdi_event')
        txt7 = "on" if parent.init_beginning else "off"
        txt6 = "on" if parent.init_sec else "off"
        txt5 = "on" if parent.init_page else "off"

        list5 = [["MDSconnect", mds_server, 204, None],
                 ["Update", gv['event'], 0, None],
                 ["Global(title) TDI",  txt4,  200, None],
                 ["Global(title) event",  txt8,  200, None],
                 ["Use global TDI",  txt3,  1, {"values": ["on", "off"]}],
                 ["Common vars", gv['shot_global'], 0, None],
                 ["Use common vars",  txt2,  1, {"values": ["on", "off"]}],
                 ["Global setting", gv['global_choice'], 21, None],
                 ["Multipage mode",  txt, 1, {"values": ["on", "off"]}],
                 #            ["Color Order", gv['color_order'], 22, {}],]
                 ]

        tmp2 = [[None,  parent.book._parallel_sec, 3,
                 {"text": 'Parallelize Panels', "noindent": True}], ]
        tmp = [[None,  (parent.book._parallel_shot, [parent.book._parallel_sec]), 27,
                ({"text": 'Parallelize Pages'}, {"elp": tmp2})]]
        list6 = [[None, "MDSplus session parallelization",  102, None],
                 [None,  (parent.book._parallel_page,
                          [(parent.book._parallel_shot, [parent.book._parallel_sec])]),
                  27,
                  ({"text": 'Parallelize Shots', "space": 15}, {"elp": tmp})],
                 #["At session start",txt7, 1, {"values":["on", "off"]}],
                 #["At every page",   txt5, 1, {"values":["on", "off"]}],
                 #["At every panel",  txt6, 1, {"values":["on", "off"]}],
                 [None, thread_main,  3,
                  {"text": "Run python script in main thread",
                   "noindent": None}],
                 [None, "(!) MDS+ connection is initialized when opening new tree",  102, None], ]
        tip1 = None
        tip2 = None
        tip3 = None
        tip4 = None
        tip5 = ['MDSplus server and default Tree',
                'Update event',
                'Global TDI expression, which runs before panels are evaulated',
                'Event for Global TDI expression',
                'Show result of global tdi in window title',
                'MDS command called at the begining of page',
                'MDS command called at the begining of section',
                'Global setting selection for shots',
                'Use the same shot numbers for all pages. (shot numbers after semi-colon are ignored)', ]

        tip6 = [None,
                'Parallelization of MDS session',
                None,
                #'Initialize global/local variables when analysis loop starts',
                #'Initialize global/local variables when analyzing a new panel',
                #'Initialize global/local variables when analyzing a new page',
                None, ]

        l = [list1, list2, list3, list4, list5, list6]
        tip = [tip1, tip2, tip3, tip4, tip5, tip6]

        self.nb = wx.Notebook(self)
        self.elp = []
        for tab, ll, t in zip(tabs, l, tip):
            self.elp.append(EditListPanel(self.nb, ll, tip=t))
            self.nb.AddPage(self.elp[-1], tab)

        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(self.nb, 1, wx.EXPAND | wx.ALL, 10)
        self.SetSizer(vbox)

        self.SetTitle('Global Setting('+parent.GetTitle().split(':')[0]+')')

        bt_apply = wx.Button(self, wx.ID_ANY, 'Apply')
        bt_save = wx.Button(self, wx.ID_ANY, 'Save')

        bsizer = GridSizer(1, 5)
        self.GetSizer().Add(bsizer, 0, wx.EXPAND | wx.ALL, 1)
        bsizer.AddStretchSpacer()
        bsizer.Add(bt_save, 1, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 3)
        bsizer.AddStretchSpacer()
        bsizer.Add(bt_apply, 1, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 3)
        bsizer.AddStretchSpacer()

        self.Bind(wx.EVT_BUTTON, self.onSave, bt_save)
#        self.Bind(wx.EVT_BUTTON, self.onCancel, bt_cancel)
        self.Bind(wx.EVT_BUTTON, self.onApply, bt_apply)
        self.Bind(wx.EVT_CLOSE, self.onClose)
        self.Layout()
        self.Show()
        # self.append_help_menu()
        # self.append_help2_menu(self.helpmenu)
        # self.set_accelerator_table()

        wx.GetApp().add_palette(self)

    def onApply(self, evt):
        value = [elp.GetValue() for elp in self.elp]
        self.cb(value)
        self.GetParent()._handle_apply_abort(allshot=True)

    def onSave(self, evt):
        value = [elp.GetValue() for elp in self.elp]
        self.cb(value)

    def onClose(self, evt):
        try:
            self.GetParent()._setting_dlg = None
        except:
            pass
        wx.GetApp().rm_palette(self)
        self.Destroy()
        evt.Skip()


def callback(parent, value, setting, gwparam):
    gs = gwparam.getvar('globalsets')
    gv = gwparam.getvar()

    if not '_flag' in gs:
        gs['_flag'] = [[], [], [], []]
    for m in range(4):
        for key in value[m][0]:
            if not key in gs:
                gs[key] = [None]*4
            gs[key][m] = value[m][0][key]
    gv['event'] = str(value[4][1])
    gv['shot_global'] = str(value[4][5])
    gv['use_shot_global'] = (value[4][6] == 'on')
    parent.book.setvar('global_tdi', str(value[4][2]))
    parent.book.setvar('global_tdi_event', str(value[4][3]))
    parent.book._use_global_tdi = value[4][4] == 'on'
    gv['global_choice'] = value[4][7]
    parent.mpanalysis = (value[4][8] == "on")

    if not value[5][1][0]:
        parent.book._parallel_page = False
        parent.book._parallel_shot = False
        parent.book._parallel_sec = False
    else:
        parent.book._parallel_page = True
        if not value[5][1][1][0][0]:
            parent.book._parallel_shot = False
            parent.book._parallel_sec = False
        else:
            parent.book._parallel_shot = True
            parent.book._parallel_sec = value[5][1][1][0][1][0]

    parent.book.setvar('mdsscript_main',
                       value[5][2])
    parent.book.setvar('mdsplus_server',
                       str(value[4][0]))

    from ifigure.utils.setting_parser import iFigureSettingParser as SettingParser
    p = SettingParser()
    v = p.read_setting('mdsplus.default_mdsserver')
    if str(v['server']) != str(value[4][0]):
        ret = dialog.message(parent,
                             '\n'.join(['Selected MDS+ server is different from your default setting',
                                        'Do you want to use this server as your default server?']),
                             'MDS+ server setting',
                             4, icon=wx.ICON_QUESTION)
#           print ret
        if ret == 'yes':
            v['server'] = str(value[4][0])
            p.write_setting('mdsplus.default_mdsserver', v)

#   def pref_setting(parent, setting):
#       pass



#
#    Margin widget
#
import wx
from ifigure.widgets.custom_double_slider import CustomDoubleSlider, EVT_CDS_CHANGED

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import UndoRedoGroupUngroupFigobj
from ifigure.widgets.undo_redo_history import UndoRedoAddRemoveArtists
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty

from ifigure.utils.wx3to4 import GridSizer, FlexGridSizer


class MarginWidget(wx.Panel):

    def __init__(self, parent, id=wx.ID_ANY, *args, **kargs):
        super(MarginWidget, self).__init__(parent, id, *args, **kargs)
        self.use_def = wx.CheckBox(self, -1, 'Use Global Setting',
                                   (10, 10))

        self.st1 = wx.StaticText(self, label="   X  ")
        self.h_slider = CustomDoubleSlider(self, wx.ID_ANY)
        self.st2 = wx.StaticText(self, label="   Y  ")
        self.v_slider = CustomDoubleSlider(self, wx.ID_ANY)

        big_sizer = wx.BoxSizer(wx.VERTICAL)
        big_sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer = FlexGridSizer(2, 2)
        sizer.AddGrowableCol(1)
        sizer.Add(self.st1, 0, wx.ALIGN_CENTER_VERTICAL, 3)
        sizer.Add(self.h_slider, 1, wx.EXPAND |
                  wx.ALIGN_CENTER_VERTICAL | wx.ALL, 3)
        sizer.Add(self.st2, 0, wx.ALIGN_CENTER_VERTICAL, 3)
        sizer.Add(self.v_slider, 1, wx.EXPAND |
                  wx.ALIGN_CENTER_VERTICAL | wx.ALL, 3)
        big_sizer.Add(self.use_def, 0)
        big_sizer.Add(sizer, 1, wx.EXPAND)
        big_sizer2.Add(big_sizer, 1, wx.EXPAND)
        self.SetSizer(big_sizer2, wx.EXPAND)

        self.Bind(EVT_CDS_CHANGED, self.onEvent)
#        self.Bind(wx.EVT_SCROLL_CHANGED, self.onEvent)
        self.Bind(wx.EVT_CHECKBOX, self.onEvent)
        self.Fit()

    def Disable(self):
        self.use_def.Disable()
        self.st1.Disable()
        self.st2.Disable()
        self.h_slider.Disable()
        self.v_slider.Disable()

    def Enable(self):
        self.use_def.Enable()
        self.st1.Enable()
        self.st2.Enable()
        self.h_slider.Enable()
        self.v_slider.Enable()

    def onEvent(self, event):
        if event.GetEventObject() is self.use_def:
            self.SetCanvasValue2()
            return
        self.SetCanvasValue()
        return

    def SetEditorValue(self, use_def, m):
        self.use_def.SetValue(use_def)
        self.h_slider.SetValue([m[0], 1.-m[1]])
        self.v_slider.SetValue([m[2], 1.-m[3]])

    def SetCanvasValue2(self, artist=None):
        a = self.h_slider.GetValue()
        b = self.v_slider.GetValue()
        c = self.use_def.IsChecked()

        se = self.GetParent().GetParent().GetParent()
        pe = se.GetParent()
        canvas = pe.get_canvas()

        f_page = (canvas.get_figure()).figobj
        if se.area_hit in se.area:
            idx = se.area.index(se.area_hit)
        else:
            se.area_hit = se.area[0]
            idx = 0

        f_axes = f_page.get_child(idx)
        if f_axes is None:
            return
        ac = [UndoRedoFigobjProperty(f_axes._artists[0],
                                     'use_def_margin', c), ]
        window = self.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(ac,
                                                       menu_name='use default margin')
        wx.CallAfter(self.GetParent().GetParent().GetParent().SetEditorValue)
        return
        f_axes.setp("use_def_margin", c)
        f_page.realize()
        canvas.draw()
        self.GetParent().GetParent().GetParent().SetEditorValue()

    def SetCanvasValue(self, artist=None):
        a = self.h_slider.GetValue()
        b = self.v_slider.GetValue()
        c = self.use_def.IsChecked()

        se = self.GetParent().GetParent().GetParent()
        pe = se.GetParent()
        canvas = pe.get_canvas()

        f_page = (canvas.get_figure()).figobj
        idx = se.area.index(se.area_hit)
        f_axes = f_page.get_child(idx)

        ac = []
        if c:
            ac.append(UndoRedoFigobjMethod(canvas.get_figure(),
                                           'def_margin',
                                           [a[0], 1.-a[1], b[0], 1.-b[1]]))
        else:
            ac.append(UndoRedoFigobjMethod(f_axes._artists[0],
                                           'margin',
                                           [a[0], 1.-a[1], b[0], 1.-b[1]]))
        window = self.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(ac,
                                                       menu_name='margin')
        wx.CallAfter(self.GetParent().GetParent().GetParent().SetEditorValue)


class MarginpWidget(wx.Panel):
    def __init__(self, parent, id=wx.ID_ANY, *args, **kargs):
        super(MarginpWidget, self).__init__(parent, id, *args, **kargs)

#        self.use_def= wx.CheckBox(self, -1, 'Use Global Setting',
#                                  (10, 10))

        self.st1 = wx.StaticText(self, label="   X  ")
        self.h_slider = CustomDoubleSlider(self, wx.ID_ANY)
        self.st2 = wx.StaticText(self, label="   Y  ")
        self.v_slider = CustomDoubleSlider(self, wx.ID_ANY)

        big_sizer = wx.BoxSizer(wx.VERTICAL)
        big_sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer = FlexGridSizer(2, 2)
        sizer.AddGrowableCol(1)
        sizer.Add(self.st1, 0, wx.ALIGN_CENTER_VERTICAL, 3)
        sizer.Add(self.h_slider, 1, wx.EXPAND |
                  wx.ALIGN_CENTER_VERTICAL | wx.ALL, 3)
        sizer.Add(self.st2, 0, wx.ALIGN_CENTER_VERTICAL, 3)
        sizer.Add(self.v_slider, 1, wx.EXPAND |
                  wx.ALIGN_CENTER_VERTICAL | wx.ALL, 3)
        big_sizer.Add(sizer, 1, wx.EXPAND)
        big_sizer2.Add(big_sizer, 1, wx.EXPAND)
        self.SetSizer(big_sizer2, wx.EXPAND)

        self.Bind(EVT_CDS_CHANGED, self.onEvent)
        self.Fit()
#        self.Bind(wx.EVT_SCROLL_CHANGED, self.onEvent)

    def Disable(self):
        self.st1.Disable()
        self.st2.Disable()
        self.h_slider.Disable()
        self.v_slider.Disable()

    def Enable(self):
        self.st1.Enable()
        self.st2.Enable()
        self.h_slider.Enable()
        self.v_slider.Enable()

    def onEvent(self, event):
        self.SetCanvasValue()

    def SetEditorValue(self, m):
        self.h_slider.SetValue([m[0], 1.-m[1]])
        self.v_slider.SetValue([m[2], 1.-m[3]])

    def SetCanvasValue(self, artist=None):
        a = self.h_slider.GetValue()
        b = self.v_slider.GetValue()

        se = self.GetParent().GetParent().GetParent()
        pe = se.GetParent()
        canvas = pe.get_canvas()

        f_page = (canvas.get_figure()).figobj
        ac = []
        ac.append(UndoRedoFigobjMethod(canvas.get_figure(),
                                       'page_margin',
                                       [a[0], 1.-a[1], b[0], 1.-b[1]]))
        window = self.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(ac,
                                                       menu_name='page margin')
        wx.CallAfter(self.GetParent().GetParent().GetParent().SetEditorValue)



from __future__ import print_function

from ifigure.utils.cbook import parseStr
from ifigure.fig_objects.py_code import PyData
from ifigure.fig_objects.py_code import PyCode
import os
import re
import string
import sys
from numpy import *
from collections import OrderedDict

######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.


#      Strong recommendation : make module "independent".

#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module
module_name = 'dwscope_subs'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
module_evt_handler = [("Load File", "onLoadFile", True)]
#
#   A function called when py_module is initialized
#
module_init = None
######################################################


def import_dwscope(file):

    f = open(file, 'r')
    while 1:
        try:
            line = f.readline()
            print(line)
        except Exception:
            print(sys.exc_info())
    f.close()


def onLoadFile(self_obj):
    open_dlg = wx.FileDialog(None, message="Select DWscope",
                             wildcard='*.dat', style=wx.FD_OPEN)
    if open_dlg.ShowModal() != wx.ID_OK:
        open_dlg.Destroy()
        return False

    file = open_dlg.GetPath()
    open_dlg.Destroy()
    for name, child in self_obj.get_children():
        child.destroy()
    self_obj.setvar("filename", file)

    import_dwscope(file)



import wx
import numpy as np
from weakref import ProxyType

isWX3 = (wx.__version__[0] == '3')

if isWX3:
    from wx._core import PyDeadObjectError
else:
    # wx4
    PyDeadObjectError = RuntimeError

if isWX3:
    from wx.aui import EVT__AUINOTEBOOK_TAB_MIDDLE_DOWN as EVT_AUINOTEBOOK_TAB_MIDDLE_DOWN
    from wx.aui import EVT__AUINOTEBOOK_TAB_RIGHT_DOWN as EVT_AUINOTEBOOK_TAB_RIGHT_DOWN
    from wx.aui import EVT__AUINOTEBOOK_TAB_RIGHT_UP as EVT_AUINOTEBOOK_TAB_RIGHT_UP
    from wx.aui import EVT__AUINOTEBOOK_TAB_MIDDLE_UP as EVT_AUINOTEBOOK_TAB_MIDDLE_UP
else:
    from wx.aui import EVT_AUINOTEBOOK_TAB_MIDDLE_DOWN
    from wx.aui import EVT_AUINOTEBOOK_TAB_RIGHT_DOWN
    from wx.aui import EVT_AUINOTEBOOK_TAB_RIGHT_UP
    from wx.aui import EVT_AUINOTEBOOK_TAB_MIDDLE_UP

if isWX3:
    from wx import TaskBarIcon as wxTaskBarIcon
    from wx import TBI_DOCK, TBI_CUSTOM_STATUSITEM, TBI_DEFAULT_TYPE
    from wx import EVT_TASKBAR_MOVE, EVT_TASKBAR_LEFT_DOWN, EVT_TASKBAR_LEFT_UP
    from wx import EVT_TASKBAR_RIGHT_DOWN, EVT_TASKBAR_RIGHT_UP, EVT_TASKBAR_LEFT_DCLICK
    from wx import EVT_TASKBAR_RIGHT_DCLICK, EVT_TASKBAR_CLICK
    from wx.combo import BitmapComboBox as wxBitmapComboBox
    from wx import IconFromBitmap
    from wx import EmptyImage as wxEmptyImage
    from wx import StockCursor as wxStockCursor
    from wx import CursorFromImage as wxCursorFromImage
    from wx import BitmapFromImage as wxBitmapFromImage
    from wx import NamedColour as wxNamedColour
    from wx import EmptyBitmapRGBA as wxEmptyBitmapRGBA
    from wx.grid import PyGridTableBase as GridTableBase

else:
    from wx.adv import TaskBarIcon as wxTaskBarIcon
    from wx.adv import TBI_DOCK, TBI_CUSTOM_STATUSITEM, TBI_DEFAULT_TYPE
    from wx.adv import EVT_TASKBAR_MOVE, EVT_TASKBAR_LEFT_DOWN, EVT_TASKBAR_LEFT_UP
    from wx.adv import EVT_TASKBAR_RIGHT_DOWN, EVT_TASKBAR_RIGHT_UP, EVT_TASKBAR_LEFT_DCLICK
    from wx.adv import EVT_TASKBAR_RIGHT_DCLICK, EVT_TASKBAR_CLICK
    from wx.adv import BitmapComboBox as wxBitmapComboBox
    from wx import Icon as IconFromBitmap
    from wx import Image as wxEmptyImage
    from wx import Cursor as wxStockCursor
    from wx import Cursor as wxCursorFromImage
    from wx import Bitmap as wxBitmapFromImage
    from wx import Colour as wxNamedColour
    wxEmptyBitmapRGBA = wx.Bitmap.FromRGBA
    from wx.grid import GridTableBase


def deref_proxy(w):
    if isWX3:
        return w
    if isinstance(w, ProxyType):
        w = w.__repr__.__self__
    return w


def wrap_method(m_wx3, m_wx4):
    def real_decorator(func):
        def wrap(obj, *args, **kwargs):
            if isWX3:
                m = getattr(obj, m_wx3)
            else:
                m = getattr(obj, m_wx4)
            args, kwargs = func(obj, *args, **kwargs)
            return m(*args, **kwargs)
        return wrap
    return real_decorator


def image2array(image):
    if isWX3:
        return np.fromstring(image.GetData(), dtype=np.uint8)
    else:
        return np.fromstring(bytes(image.GetData()), dtype=np.uint8)


def image_GetAlpha(image):
    if isWX3:
        return image.GetAlphaData()
    else:
        return image.GetAlpha()


def image_SetAlpha(image, array):
    if isWX3:
        return image.SetAlphaData(array.tobytes())
    else:
        return image.SetAlpha(array.tobytes())


def tree_InsertItemBefore(tree, pitem, pos, label, image=-1, selImage=-1, data=None):
    if isWX3:
        return tree.InsertItemBefore(pitem, pos, label, image=image)
    else:
        return tree.InsertItem(pitem, pos, label, image=image, selImage=selImage, data=data)


def grid_ClientSizeTuple(grid, *args, **kwargs):
    if isWX3:
        return grid.ClientSizeTuple()
    else:
        return grid.ClientSize


@wrap_method('SetOptionInt', 'SetOption')
def image_SetOptionInt(image, *args, **kwargs):
    return args, kwargs


@wrap_method('SetPyData', 'SetItemData')
def tree_SetItemData(tree, *args, **kwargs):
    return args, kwargs


@wrap_method('GetPyData', 'GetItemData')
def tree_GetItemData(tree, *args, **kwargs):
    return args, kwargs


@wrap_method('SetToolTipString', 'SetToolTip')
def panel_SetToolTip(panel, *args, **kwargs):
    return args, kwargs


@wrap_method('GetPositionTuple', 'GetPosition')
def evt_GetPosition(evt, *args, **kwargs):
    return args, kwargs


@wrap_method('AppendMenu', 'Append')
def menu_Append(menu, *args, **kwargs):
    return args, kwargs

menuitems = {}
def menu_AppendSubMenu(menu, *args, **kwargs):
    if isWX3:
        return menu.Append(*args, **kwargs)
    else:
        # this one replaces the following....
        #   menu_Append(helpmenu, ID_WINDOWS, 'Viewers...', self._windowmenu)
        if isinstance(args[0], wx.WindowIDRef):# and not args[0] in menuitems:
            assert False, "This does not work anymore, change your code"
            '''
            if args[0] not in menuitems:
                item = wx.MenuItem(menu, id=args[0], text=args[1], subMenu=args[2])
                menuitems[args[0]] = (item, args[2])
            else:
                item = menuitems[args[0]][0]
            return menu.Append(item)
            '''
        else:
            return menu.AppendSubMenu(args[2], args[1])

@wrap_method('AppendItem', 'Append')
def menu_AppendItem(menu, *args, **kwargs):
    return args, kwargs

@wrap_method('RemoveItem', 'Remove')
def menu_RemoveItem(menu, *args, **kwargs):
    return args, kwargs


def GridSizer(*args, **kwargs):
    if not 'hgap' in kwargs:
        kwargs['hgap'] = 0
    if not 'vgap' in kwargs:
        kwargs['vgap'] = 0
    if len(args) == 2:
        kwargs['rows'] = args[0]
        kwargs['cols'] = args[1]
    elif len(args) == 1:
        kwargs['cols'] = args[0]
    elif len(args) == 0:
        kwargs['cols'] = 1
    return wx.GridSizer(**kwargs)


def FlexGridSizer(*args, **kwargs):
    if not 'hgap' in kwargs:
        kwargs['hgap'] = 0
    if not 'vgap' in kwargs:
        kwargs['vgap'] = 0
    if len(args) == 2:
        kwargs['rows'] = args[0]
        kwargs['cols'] = args[1]
    elif len(args) == 1:
        kwargs['cols'] = args[0]
    elif len(args) == 0:
        kwargs['cols'] = 1
    return wx.FlexGridSizer(**kwargs)


def TextEntryDialog(*args, **kwargs):
    parent = args[0]
    message = args[1]
    value = kwargs.pop('value', None)
    if value is None:
        value = kwargs.pop('defaultValue', None)
    if value is None:
        value = ''

    if len(args) > 2:
        kwargs['caption'] = args[2]
    if len(args) > 3:
        value = args[3]
    if isWX3:
        kwargs['defaultValue'] = value
    else:
        kwargs['value'] = value
        parent = deref_proxy(parent)
    return wx.TextEntryDialog(parent, message, **kwargs)



import sys
import __builtin__


class RollbackImporter:
    def __init__(self):
        "Creates an instance and installs as the global importer"
        self.previousModules = sys.modules.copy()
        self.realImport = __builtin__.__import__
        __builtin__.__import__ = self._import
        self.newModules = {}

    def _import(self, name, globals=None, locals=None, fromlist=[]):
        result = self.realImport(*(name, globals, locals, fromlist))
        self.newModules[name] = 1
        return result

    def uninstall(self):
        for modname in self.newModules:
            if modname not in self.previousModules:
                # Force reload when modname next imported
                del(sys.modules[modname])
        __builtin__.__import__ = self.realImport



# evaluate cost here.
# in this skelton, cost is assumed to be evaulated
# by run script of the model and stored in param.
# Therefore it reads and retuns it.

model = args[0]
ans(model.param.getvar('cost'))



from __future__ import print_function


#  load image file
#   this could be widget toolkit dependent
import sys
import six
import os
import matplotlib.colors
from ifigure.utils.wx3to4 import menu_Append, wxBitmapFromImage, wxCursorFromImage, menu_AppendItem
from ifigure.utils.wx3to4 import image_GetAlpha, image_SetAlpha, image_SetOptionInt, wxEmptyImage
import wx
import os
import string
import weakref
import matplotlib
import threading
import traceback
import ifigure
import numpy as np
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('cbook')


try:
    import Image
except ImportError:
    from PIL import Image


def is_safename(txt):
    try:
        exec(txt + '= 3', {}, {})
    except:
        return False
    return True


def text_repr(val):
    if isinstance(val, list):
        if (len(val) > 10):
            text = (val[:10].__repr__())[:-1]+'...'
        else:
            text = val.__repr__()
    elif isinstance(val, tuple):
        if (len(val) > 10):
            text = (val[:10].__repr__())[:-1]+'...'
        else:
            text = val.__repr__()
    elif isinstance(val, dict):
        if len(val) > 5:
            text = ({key: val[key]
                     for key in list(val.keys())[0:5]}.__repr__())[:-1]+'...'
        else:
            text = val.__repr__()
    elif isinstance(val, np.ndarray):
        text = '**data**'
    elif hasattr(val, '__len__'):
        try:
            if six.PY2:
                if (len(val) > 10 and not isinstance(val, str) and not isinstance(val, unicode)):
                    text = '**data**'
                else:
                    text = val.__repr__()
            else:
                if (len(val) > 10 and not isinstance(val, str)):
                    text = '**data**'
                else:
                    text = val.__repr__()
        except:
            try:
                text = val.__repr__()
            except:
                text = ''
    else:
        try:
            text = val.__repr__()
        except:
            text = ''
    if len(text) > 40:
        text = text[0:39]+'...'
    return text


def message(*args):
    msg = ','.join([str(x) for x in args])
    print(msg)
    import ifigure.server
    s = ifigure.server.Server()
    if s.info()[0]:
        ifigure.server.Server().export_message(msg)


def get_screen_font_size(size=10, text='ABCDEF'):
    font = wx.Font(pointSize=size, family=wx.DEFAULT,
                   style=wx.NORMAL,  weight=wx.NORMAL,
                   faceName='Consolas')
    dc = wx.ScreenDC()
    dc.SetFont(font)
    return dc.GetTextExtent(text)


def get_current_display_size(window):
    '''
    get display size of current window
    support multi-screen
    '''
    scx, scy = window.ClientToScreen((0, 0))
    for i in range(wx.Display.GetCount()):
        x0, y0, xd, yd = wx.Display(i).GetGeometry()
        if (x0 < scx < x0+xd and
                scy > y0 and scy < y0+yd):
            break
    return x0, y0, xd, yd


def show_pop_up(window, menu, xy, *args, **kwargs):
    x0, y0, xd, yd = get_current_display_size(window)
    if xy[1] > yd-100:
        xy = [xy[0], yd-100]
    return window.PopupMenu(menu,  # ifigure_popup(self),
                            xy)


def isInMainThread():
    return threading.current_thread().name == 'MainThread'


def pick_unused_port():
    '''
    a routeint to le a system to pick an unused port.
    this recipe was found in active state web site
    '''
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port

#
#  conversion from PIL to WX image
#


def pil_to_image(pil, alpha=True):
    #from PIL import Image
    #import Image
    """ Method will convert PIL Image to wx.Image """
    if alpha:
        image = wxEmptyImage(*pil.size)
        image.SetData(pil.convert("RGB").tobytes())
        image.SetAlphaData(pil.convert("RGBA").tobytes()[3::4])
    else:
        image = wxEmptyImage(pil.size[0], pil.size[1])
        new_image = pil.convert('RGB')
        data = new_image.tostring()
        image.SetData(data)
    return image

#
#  conversion from WX image to PIL image
#


def image_to_pil(image):
    #
    #import Image
    """ Method will convert wx.Image to PIL Image """
    #pil = Image.new('RGB', (image.GetWidth(), image.GetHeight()))
    # pil.fromstring(image.GetData())

    data = image.GetData()

    import sys
    if isinstance(image.GetData(), bytearray):
        if sys.version_info > (3, 0):
            data = bytes(data)
        else:
            data = str(data)
    pil = Image.frombytes('RGB', (image.GetWidth(), image.GetHeight()),
                          data)
    print(pil)
    return pil


def make_crs_list(paths, hx, hy):
    crs = []
    for path in paths:
        im = wx.Image(path, wx.BITMAP_TYPE_PNG)
        # image = im.ConvertToBitmap()
        if im.HasAlpha():
            im.ConvertAlphaToMask()
        image_SetOptionInt(im, wx.IMAGE_OPTION_CUR_HOTSPOT_X, hx)
        image_SetOptionInt(im, wx.IMAGE_OPTION_CUR_HOTSPOT_Y, hy)
        crs.append(wxCursorFromImage(im))
    return crs


def make_bitmap_list(paths):
    bitmaps = []
    for path in paths:
        im = wx.Image(path, wx.BITMAP_TYPE_PNG)
        bitmaps.append(im.ConvertToBitmap())
    return bitmaps


def on_off_args(v):
    ans = False
    if isinstance(v, str):
        if v.upper() == 'ON':
            ans = True
        if v.upper() == 'OFF':
            ans = True
        if v.upper() == '1':
            ans = True
        if v.upper() == '0':
            ans = False
    elif isinstance(v, bool):
        ans = v
    elif isinstance(v, int):
        if v == 1:
            ans = True
        if v == 0:
            ans = False
    elif isinstance(v, float):
        if v == 1.0:
            ans = True
        if v == 0.0:
            ans = False
    return ans


class ImageFiles(object):
    _list = None
    _size = 16

    def __init__(self):
        if ImageFiles._list is None:
            ImageFiles._list = wx.ImageList(ImageFiles._size,
                                            ImageFiles._size)

    def add(self, fpath):
        def bmp2array(bm):
            h, w = bm.GetSize()
            im = bm.ConvertToImage()
            array = np.frombuffer(bytes(im.GetData()), dtype=np.uint8).copy()
            array = array.reshape(w, h, -1)
            alpha = image_GetAlpha(im)
            if alpha is not None:
                alpha = np.frombuffer(bytes(image_GetAlpha(im)),
                                      dtype=np.uint8).reshape(w, h, -1).copy()
            else:
                alpha = np.zeros((w, h, 1), dtype=np.uint8)+255
            return array, alpha,

        def array2bmp(array, alpha):
            w = array.shape[0]
            h = array.shape[1]
            im = wxEmptyImage(h, w)
            im.SetData(array.tobytes())
            image_SetAlpha(im, alpha)
            return im.ConvertToBitmap()

        def alpha_composite(ca, cb, aa, ab):
            resc = ca.copy().astype(float)
            resa = aa.copy().astype(float)
            ca = ca.astype(float)
            cb = cb.astype(float)
            aa = aa.astype(float)/255.
            ab = ab.astype(float)/255.
            for k in range(3):
                resc[:, :, k] = (np.multiply(ca[:, :, k], aa[:, :, 0])
                                 + np.multiply(np.multiply(cb[:, :, k], ab[:, :, 0]), 1.-aa[:, :, 0]))
            resa[:, :, 0] = aa[:, :, 0] + \
                np.multiply(ab[:, :, 0], (1.-aa[:, :, 0]))
            resa = (resa*255).astype(np.uint8)
            resc = resc.astype(np.uint8)
            return resc, resa

        if fpath[-4:] == '.png':
            bm = wxBitmapFromImage(wx.Image(fpath, wx.BITMAP_TYPE_PNG))
        if fpath[-4:] == '.bmp':
            bm = wxBitmapFromImage(wx.Image(fpath, wx.BITMAP_TYPE_BMP))

        fpath2 = os.path.join(os.path.dirname(fpath), 'suppress.png')
        bm2 = wxBitmapFromImage(wx.Image(fpath2, wx.BITMAP_TYPE_PNG))

        array, alpha = bmp2array(bm)
        array2, alpha2 = bmp2array(bm2)
        array3, alpha3 = alpha_composite(array2, array,
                                         alpha2, alpha)
        bm3 = array2bmp(array3, alpha3)
        idx = ImageFiles._list.Add(bm)
        idx2 = ImageFiles._list.Add(bm3)
        if self.IsOk(idx) is False:
            print("can not read "+fpath)
            idx = -1
            idx2 = -1
        return idx, idx2

    def get_imagelist(self):
        return ImageFiles._list

    def IsOk(self, idx):
        bm = ImageFiles._list.GetBitmap(idx)
        return bm.IsOk()

    def FilePath(self, path, file):
        if ImageFiles._size == 32:
            fpath = os.path.join(path, '32x32', file)
        if ImageFiles._size == 24:
            fpath = os.path.join(path, '24x24', file)
        if ImageFiles._size == 16:
            fpath = os.path.join(path, '16x16', file)
        return fpath


def LoadImageFile(path, fname):
    im = ImageFiles()
    fpath1 = im.FilePath(path, fname)
    return im.add(fpath1)


def Write2Main(val, name):
    import __main__
    print("writing variable "+name+" to __main__")
    exec('__main__.'+name+'=val')


def ReadFromMain(name):
    import __main__
    print("writing variable "+name+" from __main__")
    return eval('__main__.'+name)


def FindFrame(w):
    if isinstance(w, wx.Frame):
        return w
    w = w.GetParent()
    if w is not None:
        return FindFrame(w)


def GetNextName(keys, header):
    num = []
    for key in keys:
        if key.startswith(header+'_'):
            a = key[len(header)+1:]
            if a.isdigit():
                num.append(int(a))
            continue
        if key.startswith(header):
            a = key[len(header):]
            if a.isdigit():
                num.append(int(a))
            continue
    if len(num) == 0:
        return header+'1'
    else:
        return header+str(max(num)+1)


def MoveItemInList(l, i1, i2):
    # move i1 to i2
    if i1 > i2:
        return l[0:i2] + [l[i1]] + l[i2:i1] + l[i1+1:len(l)]
    elif i1 < i2:
        return l[0:i1] + l[i1+1:i2+1] + [l[i1]] + l[i2+1:len(l)]
    else:
        raise ValueError(
            "cbook::MoveItemInList :i1 should be different from i2")


def ClassNameToFile(s):
    pos = [i for i, e in enumerate(s+'A') if e.isupper()]
    parts = [s[pos[j]:pos[j+1]] for j in range(len(pos)-1)]
    return '_'.join(parts).lower()


def FileNameToClass(path):
    txt = os.path.basename(path)
    if (txt.split('.')[-1] == 'py' or
            txt.split('.')[-1] == 'pyc'):
        s = '.'.join(txt.split('.')[:-1])
    else:
        s = txt
    return ''.join([b.capitalize() for b in s.split('_')])


def isBinary(filename):
    """
    Return true if the given filename appears to be binary.
    File is considered to be binary if it contains a NULL byte.
    (This approach incorrectly reports UTF-16 as binary.)
    """
    if six.PY2:
        with open(filename, 'rb') as f:
            for block in f:
                if '\0' in block:
                    return True
    else:
        with open(filename, 'rb') as f:
            for block in f:
                if b'\0' in block:
                    return True

    return False


def isDescendant(parent, kid):
    p = kid.GetParent()
    while p is not None:
        if parent == p:
            return True
        p = p.GetParent()
    return False


def LoadClass(path, name):
    # path='ifigure.fig_objects.'
    # ClassNameToFile(name))
    import sys
#    print path+ClassNameToFile(name)
    mod = __import__(path+ClassNameToFile(name))
    return sys.modules[path+ClassNameToFile(name)]
#    components = name.split('.')
#    for comp in components[1:]:
#        comp
#        mod = getattr(mod, comp)

    return mod


def EraseBitMap(bmp):
    import array
    w = bmp.GetWidth()
    h = bmp.GetHeight()
    # Make a bitmap using an array of RGBA bytes
    bpp = 4  # bytes per pixel
    bytes = array.array('B', [0] * w*h*bpp)
    bmp.CopyFromBufferRGBA(bytes)
    return bmp


def DrawBox(box, canvas, color='red'):
    from matplotlib.lines import Line2D
    x = (box[0], box[2], box[2], box[0], box[0])
    y = (box[1], box[1], box[3], box[3], box[1])
    rb = Line2D(x, y, figure=canvas._figure,
                linestyle='-', color=color, alpha=1,
                markerfacecolor='None')
    canvas._figure.lines.extend([rb])
    canvas.draw()
    canvas._figure.lines.remove(rb)


def SetText2Clipboard(text):
    if not wx.TheClipboard.IsOpened():
        wx.TheClipboard.Open()
        data = wx.TextDataObject()
        data.SetText(text)
        wx.TheClipboard.SetData(data)
        wx.TheClipboard.Close()
        return 0
    return -1


def GetArtistExtent(a, box=None, renderer=None, canvas=None, force=False):
    '''
    find window extent of artists contained
    in a
    '''
    def merge_box(box, box2):
        return (min((box[0], box2[0])),
                min((box[1], box2[1])),
                max((box[2], box2[2])),
                max((box[3], box2[3])),)

    if box is None:
        box = (np.inf, np.inf, -np.inf, -np.inf)

    if hasattr(a, 'get_window_extent'):
        try:
            if renderer is None:
                #renderer = a.figure._cachedRenderer
                renderer = a.figure.canvas.get_renderer()

            box1 = a.get_window_extent(renderer).extents
            box = merge_box(box, box1)
        except:
            dprint1('GetArtistExtent can not check ' + str(a))
            pass
    for a in a.get_children():
        box = GetArtistExtent(a, box=box, renderer=renderer, canvas=canvas)
    if hasattr(a, 'artists'):
        for a in a.artists():
            box = GetArtistExtent(
                a, box=box, renderer=renderer, canvas=canvas, force=True)
    if hasattr(a, 'legends'):
        for a in a.legends():
            box = GetArtistExtent(
                a, box=box, renderer=renderer, canvas=canvas, force=True)
    if hasattr(a, 'get_bbox_patch'):
        if a.get_bbox_patch() is not None:
            box = GetArtistExtent(a.get_bbox_patch(), box=box,
                                  renderer=renderer, canvas=canvas, force=True)
#    box[0] = max([box[0], 0])
#    box[1] = min([w, box[1]])
#    box[2] = max([box[2], 0])
#    box[3] = min([h, box[3]])
#    box = (max((box[0], 0)),
#           max((box[1], 0)),
#           min((w, box[2])),
#           min((h, box[3])))
    if canvas:
        DrawBox(box, canvas)
    return box


def BezierFit(x, y):
    import matplotlib.path as mpath
    Path = mpath.Path

    if len(x) == 2:
        pathdata = [(Path.MOVETO, (x[0], y[0]), 0),
                    (Path.LINETO, (x[1], y[1]), 0)]
        return pathdata

    x = np.array(x)
    y = np.array(y)
    mx = (x[1:]+x[:-1])/2
    my = (y[1:]+y[:-1])/2

    cx1 = [(mx[i]-mx[i+1])/2.+x[i+1] for i in range(len(mx)-1)]
    cy1 = [(my[i]-my[i+1])/2.+y[i+1] for i in range(len(my)-1)]
    cx2 = [(mx[i+1]-mx[i])/2.+x[i+1] for i in range(len(mx)-1)]
    cy2 = [(my[i+1]-my[i])/2.+y[i+1] for i in range(len(my)-1)]

    cx2 = [(cx1[0]-x[0])/2.+x[0]]+cx2
    cy2 = [(cy1[0]-y[0])/2.+y[0]]+cy2
    cx1 = cx1+[(cx2[-1]-x[-1])/2.+x[-1]]
    cy1 = cy1+[(cy2[-1]-y[-1])/2.+y[-1]]

    pathdata = [(Path.MOVETO, (x[0], y[0]), 0)]
    for i in range(len(x)-1):
        pathdata.append((Path.CURVE4, (cx2[i], cy2[i]), 1))
        pathdata.append((Path.CURVE4, (cx1[i], cy1[i]), 1))
        pathdata.append((Path.CURVE4, (x[i+1], y[i+1]), 1))
    return pathdata


def BezierComputeCurve(pathdata, mesh=10):
    def numerical_cubic_bezier_fn(control_pts, t):
        # define the actual cubic bezier equation here
        def fn(c, t): return c[0]*(1 - t)**3 + c[1]*3 * \
            t*(1 - t)**2 + c[2]*3*t**2*(1 - t) + c[3]*t**3
        xs = [x for x, y in control_pts]
        ys = [y for x, y in control_pts]
        # now calculate the x,y position from the bezier equation
        xpt = fn(xs, t)
        ypt = fn(ys, t)
        return xpt, ypt
    import matplotlib.path as mpath
    Path = mpath.Path
    t = np.linspace(0, 1, mesh)[:-1]
    ii = iter(pathdata)
    x = []
    y = []
    lastitem = next(ii)
    while 1:
        try:
            item1 = next(ii)
        except StopIteration:
            break
        if item1[0] == Path.MOVETO:
            lastitem = item[1]
        elif item1[0] == Path.CURVE4:
            item2 = next(ii)
            item3 = next(ii)
            cps = [lastitem[1], item1[1], item2[1], item3[1]]
            a, b = numerical_cubic_bezier_fn(cps, t)
            x.extend(a)
            y.extend(b)
            lastitem = item3
    return x, y


def BezierNodeType(path, inode):
    import matplotlib.path as mpath
    Path = mpath.Path

    type = 1
    j = 0
    '''
    node on the line is
    type = 1, 2, 5
    '''
    for i in range(len(path)):
        if (path[i][0] == Path.MOVETO or
                path[i][0] == Path.LINETO):
            type = 1
            j = 0
        elif path[i][0] == Path.CURVE4:
            j = j+1
            if j % 3 == 0:
                type = 2
            elif j % 3 == 1:
                type = 3
            else:
                type = 4
        elif path[i][0] == Path.POLYCLOSE:
            type = 5
        if i == inode:
            return type
    return type


def BezierRmnode(path, hit):
    '''
    remove control hit-th node from path
    '''
    import matplotlib.path as mpath
    Path = mpath.Path
#   print 'removing node ', hit
    segpath = BezierSplit(path)
    if len(segpath) == 1:
        return None
    if hit == 0:
        segpath = segpath[1:]
    else:
        i = 0
        iseg = 0
        for seg in segpath:
            #          print 'checking', i+len(seg), hit
            if i+len(seg)-1 >= hit:
                break
            i = i+len(seg)-1
            iseg = iseg+1
        iseg = min([iseg, len(segpath)])
#      print 'merging segment', iseg
        if iseg == len(segpath)-1:
            segpath = segpath[:-1]
        else:
            #          print 'merging segment', iseg, iseg+1
            seg1 = segpath[iseg]
            seg2 = segpath[iseg+1]
            # iseg segment which hit
            if (len(seg1) == 4 and len(seg2) == 4):  # curve-curve
                newseg = [seg1[0], seg1[1], seg2[2], seg2[3]]
            elif (len(seg1) == 4 and len(seg2) == 2):  # curve-line
                newseg = [seg1[0], seg2[1]]
            elif (len(seg1) == 2 and len(seg2) == 4):  # line-curve
                seg2m = (Path.LINETO, seg2[3][1], seg2[3][2])
                newseg = [seg1[0], seg2[3]]
            elif (len(seg1) == 2 and len(seg2) == 2):  # line-line
                newseg = [seg1[0], seg2[1]]

            segpath = segpath[:iseg]+[newseg]+segpath[iseg+2:]
    p = BezierJoin(segpath)
    return p


def BezierSplit(path):
    import matplotlib.path as mpath
    Path = mpath.Path

    segs = []
    seg = []

    i = 0
    while i < len(path):
        if path[i][0] == Path.MOVETO:
            seg = [path[i]]
        elif path[i][0] == Path.LINETO:
            seg = seg + [path[i]]
            segs.append(seg)
            seg = [(Path.MOVETO, path[i][1], 0)]
        elif path[i][0] == Path.CURVE4:
            seg = seg + [path[i]]
            i = i+1
            seg = seg + [path[i]]
            i = i+1
            seg = seg + [path[i]]
            segs.append(seg)
            seg = [(Path.MOVETO, path[i][1], 0)]
        i = i+1

    return segs


def BezierInsert(path, i,  x, y):
    segpath = BezierSplit(path)
    segpath[i] = BezierInsert0(segpath[i],  x, y)
    path = BezierJoin(segpath)
    return path


def BezierJoin(segpath):
    path = [segpath[0][0]]
    for seg in segpath:
        path.extend(seg[1:])
    return path


def BezierInsert0(segpath,  x, y):
    import matplotlib.path as mpath
    Path = mpath.Path

    st = segpath[0][1]
    if segpath[-1][0] == Path.LINETO:
        ans = [segpath[0],
               (Path.CURVE4, (int(st[0]*0.3+x*.7), int(st[1]*0.3+y*.7)), 2),
               (Path.CURVE4, (x, y), 2), ]
        st = ans[-1][1]
        x, y = segpath[-1][1]
        ans = ans + [
            (Path.CURVE4, (int(st[0]*0.7+x*.3), int(st[1]*0.7+y*.3)), 2),
            segpath[-1]]
    elif segpath[-1][0] == Path.CURVE4:
        ans = [segpath[0],
               segpath[1],
               (Path.CURVE4, (int(st[0]*0.3+x*.7), int(st[1]*0.3+y*.7)), 2),
               (Path.CURVE4, (x, y), 2)]
        st = ans[-1][1]
        x, y = segpath[-1][1]
        ans = ans + [
            (Path.CURVE4, (int(st[0]*0.7+x*.3), int(st[1]*0.7+y*.3)), 2),
            segpath[-2],
            segpath[-1]]
    return ans


def BezierHitTest(path, x0, y0):
    from matplotlib.patches import PathPatch
    import ifigure.widgets.canvas.custom_picker as cpicker

    segpath = BezierSplit(path)
    i = 0
    for seg in segpath:
        p = [(item[0], item[1]) for item in seg]
        codes, verts = zip(*p)
        obj = matplotlib.path.Path(verts, codes)
        a = PathPatch(obj)
        xy = a.get_verts()
        if len(xy) == 0:  # this case happens when verts has only two points in mpl1.5
            x = [v[0] for v in verts]
            y = [v[1] for v in verts]
        else:
            x = np.transpose(xy)[0]
            y = np.transpose(xy)[1]
        hit, idx = cpicker.CheckLineHit(x, y, x0, y0)
        if hit:
            return True, i
        i = i+1
    return False, -1


def ParseNameObj(*argc):
    name = None
    obj = None
    if len(argc) == 2:
        name = argc[0]
        obj = argc[1]
    elif len(argc) == 1:
        if type(argc[0]) is str:
            name = argc[0]
        else:
            obj = argc[0]
    return name, obj


def BuildPopUpMenu(base, menus, eventobj=None,
                   xy=None, xydata=None):
    '''    
     base = wx.Menu
     menus = [(menu_string, func, icon_image)]

     special menu_string
       --- : separator
       '+' and '!' control submenu generation
        +...:  goto deeper menu
        !...: end of current depth

     icon_image is not yet implemented...

     if eventobj is set to widget, events returned
     from the popup has the widget as eventobj
    '''
    isTop = True
    ret = {}
    for m in menus:
        s = m[0]
        h = m[1]
        i = m[2]
        bmp = None if len(m) < 4 else m[3]
        id = wx.ID_ANY if len(m) < 5 else m[4]
        if s == '---':
            if not isTop:
                base.AppendSeparator()
            continue
        elif s[0] == '+':
            new_base = wx.Menu()
            mmi = base.AppendSubMenu(new_base, s[1:])
            base = new_base
            isTop = True
            mmm = base
        elif s[0] == '!':
            base = base.GetParent()
            isTop = False
        else:
            isTop = False
            escape = False
            if s[0] == "\\":
                escape = True
                mmi = wx.MenuItem(base, id, s[1:])
            elif s[0] == '-':
                mmi = wx.MenuItem(base, id, s[1:])
            elif s[0] == '*':
                mmi = wx.MenuItem(base, id, s[1:], kind=wx.ITEM_CHECK)
            elif s[0] == '^':
                mmi = wx.MenuItem(base, id, s[1:], kind=wx.ITEM_CHECK)
            else:
                mmi = wx.MenuItem(base, id, s)
            if bmp is not None:
                mmi.SetBitmap(bmp)
            menu_AppendItem(base, mmi)
            if not escape:
                if s[0] == '^':
                    mmi.Check(True)
                if s[0] == '-':
                    mmi.Enable(False)

            def func(evt, handler=h, obj=eventobj, extra=i,
                     xy=xy, xydata=xydata):
                if obj is not None:
                    evt.SetEventObject(obj)
                if extra is not None:
                    evt.ExtraInfo = extra
                if xy is not None:
                    evt.mpl_xy = xy
                else:
                    evt.mpl_xy = (None, None)
                if xydata is not None:
                    evt.mpl_xydata = xydata
                else:
                    evt.mpl_xydata = (None, None)
                handler(evt)
            base.Bind(wx.EVT_MENU, func, mmi)
            mmm = mmi.GetMenu()
        if id != wx.ID_ANY:
            ret[id] = (mmm, mmi)
    return ret


BuildMenu = BuildPopUpMenu


def parseStr(x0):
    #
    # convert string to number
    #   3D3, 3d3, 3E3, 3e3 to float
    #   the bottom part wad downloaded from somewhere
    #

    num = 1
    if x0.find('*') != -1:
        x = x0.split('*')[1]
        num = int(x0.split('*')[0])
    else:
        x = x0

    sign = 1
    x1 = x
    if x.startswith('+'):
        x1 = x[1:]
        sign = 1
    if x.startswith('-'):
        x1 = x[1:]
        sign = -1
    try:
        if x1.isdigit():
            return [int(x1)*sign]*num
    except:
        pass

    try:
        return [float(x1)*sign]*num
    except:
        pass

    try:
        return [float(x1.replace('D', 'E'))*sign]*num
    except:
        pass

    try:
        return [float(x1.replace('d', 'E'))*sign]*num
    except:
        pass

    try:
        return [x.isalpha() and x or x.isalnum() and x or len(set(string.punctuation).intersection(x)) == 1 and x.count('.') == 1 and float(x) or x]*num
    except:
        pass

    return [None]


def isdynamic(t):
    if (isinstance(t, str) and
        t.startswith('=') and
            len(t) != 1):
        return True
    else:
        return False


def isiterable(obj):
    'return true if *obj* is iterable'
    try:
        iter(obj)
    except TypeError:
        return False
    return True


def isiterable_not_string(obj):
    return hasattr(obj, '__iter__') and not isinstance(obj, str)


def isstringlike(x):
    if six.PY2:
        return (isinstance(x, str) or isinstance(x, unicode))
    else:
        return isinstance(x, str)


def nd_iter(x):
    if x.size:
        return x
    return []


def issequence(obj):
    'return true if *obj* is iterable'
    try:
        obj[0:0]
    except (AttributeError, TypeError):
        return False
    return True


def isnumber(obj):
    '''
    this may not catch all of strange classes....
    this also does not work with bool
    '''
    attrs = ['__add__', '__sub__', '__mul__', '__truediv__', '__pow__']
    return all(hasattr(obj, attr) for attr in attrs) and not (isinstance(obj, np.ndarray) and obj.dtype.kind in 'OSU')


def isndarray(obj):
    return isinstance(obj, np.ndarray)


def GetModuleDir(mname):
    import sys
    namelist = mname.split('.')
    if (namelist[0] in sys.modules) is False:
        try:
            top = __import__(namelist[0], globals(), locals(), [], -1)
        except Exception:
            return ''
    p0 = sys.modules[namelist[0]].__path__[0]
    if len(namelist) == 1:
        return p0
    k = 1
    for name in namelist[1:]:
        p0 = os.path.join(p0, name)
    return p0


def LoadScriptFile(file):
    import os
    import time
    import logging
    logging.basicConfig(level=logging.DEBUG)

    mtime = os.path.getmtime(file)
    f = open(file, 'r')
    txt = ''
    while 1:
        line = f.readline()
        if not line:
            break
        txt = txt+line
    f.close()

    try:
        code = compile(txt, file, 'exec')
    except:
        print(traceback.format_exc())
        raise

    return txt, code, mtime


def WeakNone():
    class tmp(object):
        pass
    x = tmp()
    ref = weakref.ref(x)
    del x
    return ref


def ProxyAlive(wp):
    if wp is None:
        return False
    try:
        wp.ref()
        return True
    except Exception:
        return False


def LaunchEmacs(file):
    import wx

    if wx.Platform == '__WXMAC__':
        txt = 'open -a /Applications/Emacs.app ' + file
    else:
        txt = 'emacs '+file + ' &'
    os.system(txt)


def ProcessKeywords(kywds, name, value=None):
    if name in kywds:
        value = kywds[name]
        del kywds[name]
    return value, kywds


def test(a=None, b=None, c=None):
    print(a)
    print(b)
    print(c)


try:
    from scipy import genfromtxt as genfromtxt
except ImportError:
    from numpy import genfromtxt as genfromtxt


def loadct(num, **kwargs):
    file = os.path.join(ifigure.__path__[0], 'resources', 'idl_colors.txt',)

    output = genfromtxt(file,
                        skip_header=256*num,
                        skip_footer=(39-num)*256)/255.
    return matplotlib.colors.LinearSegmentedColormap.from_list('idl'+str(num),
                                                               output, **kwargs)


def register_idl_colormaps():
    import matplotlib as mpl
    names = []
    for x in range(40):
        cmap = loadct(x)
        name = 'idl'+str(x)
        mpl.colormaps.register(cmap, name=name)
        names.append(name)
    return names


def tex_escape(text):
    """
        :param text: a plain text message
        :return: the message escaped to appear correctly in LaTeX
    """
    import re
    conv = {
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\^{}',
        '\\': r'\textbackslash{}',
        '<': r'\textless',
        '>': r'\textgreater',
    }
    if six.PY2:
        regex = re.compile('|'.join(re.escape(unicode(key))
                                    for key in sorted(list(conv.keys()), key=lambda item: - len(item))))
    else:
        regex = re.compile('|'.join(re.escape(key)
                                    for key in sorted(list(conv.keys()), key=lambda item: - len(item))))

    return regex.sub(lambda match: conv[match.group()], text)


def escape_split(s, delim):
    ret = []
    current = []
    itr = iter(s)
    for ch in itr:
        if ch == '\\':
            try:
                # skip the next character; it has been escaped!
                current.append('\\')
                current.append(next(itr))
            except StopIteration:
                pass
        elif ch == delim:
            # split! (add current to the list and reset it)
            ret.append(''.join(current))
            current = []
        else:
            current.append(ch)
    ret.append(''.join(current))
    return ret


def tex_escape_equation(text):
    arr = escape_split(text, '$')
#    print arr
#    if arr[0] != '':
#        arr2 = [x if i%2 else tex_escape(x) for i,x in enumerate(arr)]
#    else:
#        arr2 = [x if not i%2 else tex_escape(x) for i,x in enumerate(arr)]
    arr2 = [x if i % 2 else tex_escape(x) for i, x in enumerate(arr)]
    delim = ['$']*(len(arr)-1)
    if len(delim) % 2:
        delim[-1] = r'\$'
    return ''.join([x+y for x, y in zip(arr2, delim+[''])])
#    if arr[0] != '':
#       return ''.join([x+y for x, y in zip(arr2, delim+[''])])
#    else:
#        return ''.join([y+x for x, y in zip(arr2, delim+[''])])
#    if arr[0] == '': arr[0] = '\\'
#    return '$'.join(arr)


def walk_OD_tree(od, basekey=''):
    '''
    walk OrdereDictionary tree
    '''
    import collections
    for key in list(od.keys()):
        if isinstance(od[key], collections.OrderedDict):
            if basekey == '':
                basekey2 = key
            else:
                basekey2 = basekey+'.'+key
            for key2 in walk_OD_tree(od[key], basekey=basekey2):
                #               yield key2, data
                yield key2
        else:
            if basekey == '':
                basekey2 = key
            else:
                basekey2 = basekey+'.'+key
#           yield basekey2, od[key]
            yield basekey2


def read_OD_tree(od, keys):
    arr = keys.split('.')
    ret = od
    for k in arr:
        ret = ret[k]
    return ret


class DictDiffer(object):
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict, past_dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = set(
            current_dict.keys()), set(past_dict.keys())
        self.intersect = self.set_current.intersection(self.set_past)

        self.show_detail()

    def show_detail(self):
        print(("Added:", self.added()))
        for item in self.added():
            print((item, ':', self.current_dict[item]))

        print(("Removed:", self.removed()))
        for item in self.removed():
            print((item, ':', self.past_dict[item]))

        print(("Changed:", self.changed()))
        for item in self.changed():
            print((item, ':', self.current_dict[item], self.past_dict[item]))

        print(("Unchanged:", self.unchanged()))

    def added(self):
        return self.set_current - self.intersect

    def removed(self):
        return self.set_past - self.intersect

    def changed(self):
        return set(o for o in self.intersect if self.past_dict[o] != self.current_dict[o])

    def unchanged(self):
        return set(o for o in self.intersect if self.past_dict[o] == self.current_dict[o])



from matplotlib.collections import LineCollection
from matplotlib.cm import ScalarMappable
import numpy as np
from matplotlib.artist import allow_rasterization
from matplotlib.path import Path
from matplotlib.transforms import Bbox, TransformedPath
from matplotlib.colors import Colormap


class CZLineCollection(LineCollection):
    def __init__(self,  x, y, cz,
                 cmap='jet', linestyle='solid',
                 pickradius=5,  alpha=None,  **kargs):
        if linestyle is None:
            kargs['linestyle'] = 'solid'
            self._nodraw = True
        else:
            kargs['linestyle'] = linestyle
            self._nodraw = False
        self.x = x
        self.y = y
        self.cz = cz
        self.set_xydata()
        self.pickradius = pickradius
        segments = self._calc_segments()
#        self.set_linewidth(1)
        LineCollection.__init__(self, segments, **kargs)
        self.set_array(cz)
        self._cz_linesytle_name = linestyle
#        self._cz_alpha = alpha
        self._transformed_path = None
        self.set_alpha(alpha)

    def get_xdata(self):
        return self.x

    def get_ydata(self):
        return self.y

    def get_czdata(self):
        return self.cz

    def set_xydata(self):
        self._xydata = np.transpose(np.array([self.x, self.y]))

    def set_xdata(self, x):
        self.x = x
        if (x.size != self.y.size or
                x.size != self.cz.size):
            return
        segments = self._calc_segments()
        self.set_segments(segments)
        self.set_xydata()
        self.set_array(self.cz)
        self.set_alpha(self._cz_alpha)

    def set_ydata(self, y):
        self.y = y
        if (y.size != self.x.size or
                y.size != self.cz.size):
            return
        segments = self._calc_segments()
        self.set_segments(segments)
        self.set_xydata()
        self.set_array(self.cz)
        self.set_alpha(self._cz_alpha)

    def set_czdata(self, cz):
        self.cz = cz
        if (cz.size != self.x.size or
                cz.size != self.y.size):
            return
        segments = self._calc_segments()
        self.set_segments(segments)
        self.set_xydata()
        self.set_array(self.cz)
        self.set_alpha(self._cz_alpha)

    def set_alpha(self, v):
        #        LineCollection.set_alpha(self, v)
        #        calling superclass set_alpha does make color along
        #        the line constant
        self._alpha = None
        self._cz_alpha = v
        ec = self.get_edgecolors()
        r = (ec[:, 0] != 0.)
        g = (ec[:, 1] != 0.)
        b = (ec[:, 2] != 0.)
        if self._cz_alpha is None:
            v = 1.0
        ec[np.logical_or(np.logical_or(r, g), b), 3] = v

#        self.get_edgecolors()[:,3] = v
    def update_scalarmappable(self):
        #
        #  update_scalamappable in super class will
        #  set proper size of edgecolors
        #  then I need to set alpha
        #
        LineCollection.update_scalarmappable(self)
        self.set_alpha(self._cz_alpha)

    def get_alpha(self):
        return self._cz_alpha

    def set_color(self, v):
        pass

    def get_color(self):
        pass

    def set_linewidth(self, v):
        LineCollection.set_linewidth(self, v)

    def get_linewidth(self):
        v = LineCollection.get_linewidth(self)[0]
        return v

    def set_linestyle(self, v):
        if v is None or v == 'None':
            self._nodraw = True
        else:
            self._nodraw = False
            LineCollection.set_linestyle(self, v)
        self._cz_linesytle_name = v

    def get_linestyle(self):
        if self._nodraw:
            return 'None'
        return self._cz_linesytle_name

    def set_marker(self, v):
        pass

    def get_marker(self):
        pass

    def set_markersize(self, v):
        pass

    def get_markersize(self):
        pass

    def set_markeredgecolor(self, v):
        pass

    def get_markeredgecolor(self):
        pass

    def set_markerfacecolor(self, v):
        pass

    def get_markerfacecolor(self):
        pass

    def set_markeredgewidth(self, v):
        pass

    def get_markeredgewidth(self):
        pass

    def contains(self, mouseevent):
        '''
        a hit test based on what is used in 
        line2D
        '''
        from matplotlib.lines import segment_hits
        if self.figure is None:
            pixels = self.pickradius
        else:
            pixels = self.figure.dpi / 72. * self.pickradius

        olderrflags = np.seterr(all='ignore')
        transformed_path = self._get_transformed_path()
        path, affine = transformed_path.get_transformed_path_and_affine()
        path = affine.transform_path(path)
        xy = path.vertices
        xt = xy[:, 0]
        yt = xy[:, 1]

        try:
            # If line, return the nearby segment(s)
            ind = segment_hits(mouseevent.x, mouseevent.y, xt, yt, pixels)
        finally:
            np.seterr(**olderrflags)
        return len(ind) > 0, dict(ind=ind)

    def get_window_extent(self, renderer):
        bbox = Bbox([[0, 0], [0, 0]])
        trans_data_to_xy = self.get_transform().transform
        bbox.update_from_data_xy(trans_data_to_xy(self._xydata),
                                 ignore=True)
        return bbox

    @allow_rasterization
    def draw(self, *argrs, **kargs):
        self._transform_path()
        if not self._nodraw:
            LineCollection.draw(self, *argrs, **kargs)

    def _calc_segments(self):
        points = np.array([self.x, self.y]).T.reshape(-1, 1, 2)
        return np.concatenate([points[:-1], points[1:]], axis=1)

    def _transform_path(self):
        _path = Path(self._xydata)
        self._transformed_path = TransformedPath(_path, self.get_transform())

    def _get_transformed_path(self):
        """
        Return the :class:`~matplotlib.transforms.TransformedPath` instance
        of this line.
        """
        if self._transformed_path is None:
            self._transform_path()
        return self._transformed_path



from __future__ import print_function
import numpy as np
import collections
from scipy.io.idl import readsav
import wx
import os
import traceback
import shutil
import ifigure
from ifigure.mto.py_contents import IDLfile, IDLData
from ifigure.utils.recarray import rec2dict
######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.
#      Strong recommendation : make module "independent".
#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module/class
module_name = 'idl_savfile'
class_name = 'idl_savfile'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("Import...",   "onLoadFile", True),
        #      ("Export...",   "onExportFile", True),
        #      ("Update File",   "onUpdateFile", True),
        ("Update Tree",   "onUpdateTree", True)]
method = ['onLoadFile',
          'onUpdateTree',
          'init', 'init_after_load']

icon = 'data.png'

modename = 'idl_pathmode'
pathname = 'idl_path'
extname = 'idl_ext'
wildcard = 'idl(*.sav)|*.sav|Any|*'
######## Do not chage this field ####################
can_have_child = False
has_private_owndir = False
######################################################


def init(self, *args, **kargs):
    #   a function called when py_module is initialized
    obj = self.td
    obj.mk_owndir()
    nm = IDLfile()  # nm['data'] = IDLData()
    obj.setvar0(nm)
    if 'src' not in kargs:
        self.onLoadFile(None)
    # self.onLoadFile()


def load_matfile(obj):
    file = obj.path2fullpath(modename=modename,
                             pathname=pathname)
    if file == '':
        return
    # clean first
    for name, child in obj.get_children():
        child.destroy()

    def idl2dict(dd, cls=collections.OrderedDict):
        r = cls()
        for name in dd:
            print(name)
            print(isinstance(dd[name], np.recarray))
            if isinstance(dd[name], np.recarray):
                r[name] = rec2dict(dd[name], cls=cls)
            else:
                r._var0[name] = dd[name]
        return r

    nm0 = readsav(file)
    if nm0 is None:
        nm0 = IDLfile()
    else:
        nm0 = idl2dict(nm0, cls=IDLfile)
    nm = IDLfile()
    nm['data'] = nm0
    obj.setvar0(nm)

#   nm = IDLfile(nm0);nm['data'] = IDLData()
#   if nm0 is None: return
#   for key in nm0: nm['data'][key] = nm0[key]


def export_matfile(obj, file):
    nm = obj.getvar0()
    nm = sio.savemat(file, nm['data'])


def onUpdateTree(self, e):
    obj = self.td
    load_matfile(obj)


def onLoadFile(self, e):
    from ifigure.utils.addon_utils import onLoadFile

    ret = onLoadFile(self.td, message="Select .sav File",
                     modename=modename,
                     pathname=pathname,
                     extname=extname,
                     wildcard=wildcard,
                     ask_org_copy=False)
    if ret:
        load_matfile(self.td)
        ifigure.events.SendChangedEvent(self.td)


def init_after_load(self, olist, nlist):
    obj = self.td
    load_matfile(obj)



from __future__ import print_function
import wx
import ifigure
from ifigure.utils.cbook import parseStr
from ifigure.mto.py_code import PyData
from ifigure.mto.py_code import PyCode
from ifigure.mto.py_module import PyModule
from ifigure.mto.py_contents import MDSPlusTree
#from ifigure.add_on.data.mdsplus_tree import MdsplusTreeNode
import wx
import sys
import os
import re
import string
from numpy import *
from collections import OrderedDict
#import ifigure.utils.mdsplusr as mds
from ifigure.utils.edit_list import DialogEditList
from ifigure.mdsplus.mdsscope import print_threaderror

######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.


#      Strong recommendation : make module "independent".

#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module
module_name = 'mdsplus_tree'
class_name = 'mdsplus_tree'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = [("Open Tree", "onOpenTree", True)]

#
#   method lists module functions which will
#   be registered as method
#
#   spceical methods:
#      init  : called when mto will be made and
#              this module is first loaded.
#      clean : called when mto will be killed
#
method = ['onOpenTree', 'OpenTreeDone', 'init']

icon = 'data.png'
can_have_child = True
has_private_owndir = True
######################################################


def split_str(s, pattern):
    ret = re.split(pattern, s)

    ret2 = []
    for s in ret:
        s = s.strip()
        if s != '':
            ret2 = ret2+[s]
    return ret2


def make_treedicts(tree, this):
    arr = split_str(tree[0], r'[\:.]')

    treename = arr[0][1:]

    top = PyData()
    top.set_can_have_child(False)
    this.add_child(treename, top)
    dicttop = MDSPlusTree()
    top.setvar0(dicttop)

    print(("tree nodes :", len(tree)))

    for k in range(len(tree)-1):
        arr = split_str(tree[k+1], r'[\:.]')
        p = dicttop
        for key in arr[1:]:
            p.setdefault(key, MDSPlusTree())
            p = p[key]

        p.mds_path = tree[k+1]
    return top


def ask_mdssetting(this=None):
    if this is not None:
        server = this.getvar("server")
        port = str(this.getvar("port"))
    else:
        server = "localhost"
        port = "10002"
    tree = "ANALYSIS"
    shot = "-1"
    list = [["", " " * 50, 2],
            #           ["Server", server, 0],
            #           ["Port", port, 0],
            ["MDSconnect", this.getvar('mdsplus_server'), 204, None],
            ["Tree", tree, 4, {"choices": ["ANALYSIS", "ELECTRONS",
                                           "SPECTROSCOPY", "XTOMO", "DNB", "RF",
                                           ]}],
            ["Shot", shot, 0]]

    flag, value = DialogEditList(list, parent=this.get_app())

    if flag:
        this.setvar('mdsplus_server', str(value[1]))
        this.setvar("tree", str(value[2]))
        this.setvar("shot", str(value[3]))
    return flag, value


def init(self, src=None):
    td = self.td
    proj = self.td.get_root_parent()
    if proj.setting.has_child('mdsplus_worker'):
        workers = proj.setting.mdsplus_worker
    else:
        file = os.path.join(ifigure.__path__[0], 'add_on',
                            'setting', 'module', 'mdsplus_worker.py')

        workers = proj.setting.add_absmodule(file)
        workers.rename('mdsplus_worker')
    workers.call_method('onStartWorker')
    if not td.hasvar('mdsplus_server'):
        td.setvar('mdsplus_server', 'direct::CMOD')

#    if src is None:
#       flag, value = ask_mdssetting()
#       td._name='mdsplus'
#       self.onOpenTree(tree =value[1],
#                       shot =long(value[2]))


def OpenTreeDone(self, ana):

    if not 'list' in ana.result:
        return
    this = self.td
    treelist = ana.result['list']
    if treelist is None:
        return

    top = make_treedicts(treelist, this)
    top.setvar("tree", ana.tree)
    top.setvar("shot", ana.shot)
    ifigure.events.SendChangedEvent(self.td,
                                    useProcessEvent=False)


def onOpenTree(self, ev=None,
               server=None,
               port=None,
               tree=None,
               shot=None,):
    this = self.td
    if (server is None and
        port is None and
        tree is None and
            shot is None):
        # this is when "open tree" was chosen
        # in popup menu of varviwer
        flag, value = ask_mdssetting(this)
        if flag is False:
            return

    if server is None:
        server = this.getvar('mdsplus_server')
    if tree is None:
        tree = this.getvar("tree")
    if shot is None:
        shot = int(this.getvar("shot"))

    from ifigure.mdsplus.fig_mds import MDSsession
    from ifigure.mdsplus.mds_job import MDSjob
    from ifigure.mdsplus.mdsscope import start_mds_threads, message

    ana = MDSsession()
    job0 = MDSjob('connection_mode', server)
    job1 = MDSjob('open', tree, shot)
    job2 = MDSjob('value', 'getnci("***", "FULLPATH")')
    ana.add_job([job0], 'connection_mode')
    ana.add_job([job1], 'connection')
    ana.add_job([job2], 'list')
    ana.isec = 0
    ana.ishot = 0
    ana.ipage = 0
    ana.shot = shot
    ana.do_title = False
    ana.tree = tree

    sr = start_mds_threads()  # sr = session_runner
    c = message('run', ([ana], wx.GetApp().TopWindow,
                        self.OpenTreeDone, False))
    sr.queue.put(c)



from __future__ import print_function
'''
   a routin to expand goemetry array

   equivalent to the following IDL

   xvctr=mdsvalue('\analysis::top.limiters.tiles:XTILE')
   yvctr=mdsvalue('\analysis::top.limiters.tiles:YTILE')
   nvctr=mdsvalue('\analysis::top.limiters.tiles:NSEG')
   lvctr=mdsvalue('\analysis::top.limiters.tiles:PTS_PER_SEG')

   for i=0,nvctr-1 do begin
     x = n_elements(x) eq 0 ? xvctr(0:lvctr(i)-1,i) $
       : [x, !values.f_nan, xvctr(0:lvctr(i)-1,i)]
     y = n_elements(y) eq 0 ? yvctr(0:lvctr(i)-1,i) $
       : [y, !values.f_nan, yvctr(0:lvctr(i)-1,i)]   
   end
'''
import numpy as np


def expand_vctr(x, y, n, l):
    print(x.shape, end=' ')
    res_x = [np.hstack((x[k, :l[k]], np.nan)) for k in range(n)]
    res_y = [np.hstack((y[k, :l[k]], np.nan)) for k in range(n)]

    return np.hstack(res_x)[:-2], np.hstack(res_y)[:-2]


def expand_vctr3d(xyz, n, l, list=False):
    if list:
        res_x = [np.hstack(xyz[k, :l[k], 0]) for k in range(n)]
        res_y = [np.hstack(xyz[k, :l[k], 1]) for k in range(n)]
        res_z = [np.hstack(xyz[k, :l[k], 2]) for k in range(n)]
        return res_x, res_y, res_z
    else:
        res_x = [np.hstack((xyz[k, :l[k], 0], np.nan)) for k in range(n)]
        res_y = [np.hstack((xyz[k, :l[k], 1], np.nan)) for k in range(n)]
        res_z = [np.hstack((xyz[k, :l[k], 2], np.nan)) for k in range(n)]
        return np.hstack(res_x)[:-2], np.hstack(res_y)[:-2], np.hstack(res_z)[:-2]


def read_vtcr(file):
    from scipy.io.idl import readsav
    nm0 = readsav(file)
    return expand_vctr(nm0['xvctr'], nm0['yvctr'],
                       nm0['nvctr'], nm0['lvctr'])



from packaging.version import parse as parse_version
import matplotlib as mpl
mpl_version = mpl.__version__


def get_color_cycle(axes):
    if parse_version(mpl_version) >= parse_version('1.5'):
        return axes._get_lines.prop_cycler
    else:
        return axes._get_lines.color_cycle


def reset_color_cycle(axes):
    if parse_version(mpl_version) >= parse_version('1.5'):
        axes.set_prop_cycle(None)  # this reset color cycle
    else:
        pass
        #cycle = get_color_cycle(axes)
        #i = 0
        # while i < 50:
        #    c = cycle.next()
        #    if c == color_cycle[-1]: break
        #    i = i + 1


def get_color_cycle_list(axes):
    if parse_version(mpl_version) >= parse_version('1.5'):
        return [x['color'] for x in mpl.rcParams['axes.prop_cycle']]
    else:
        import six
        color = [six.next(get_color_cycle(axes))
                 for i in range(nx)]


def call_savefig_method(ifigure_canvas, name, *args, **kargs):
    '''
    print_pdf, and so on has gone in version 1.4
    this function is to absorb this issue.
    '''
    canvas = ifigure_canvas.canvas
    figure = ifigure_canvas._figure
    if hasattr(canvas, name):
        m = getattr(canvas, name)
        return m(*args, **kargs)
    else:
        return figure.savefig(*args, **kargs)



from __future__ import print_function
# uncomment the following to use wx rather than wxagg
from matplotlib.backends.backend_agg import RendererAgg
import os
from ifigure.matplotlib_mod.backend_wxagg_mod import FigureCanvasWxAggMod
from distutils.version import LooseVersion
import numpy as np
import time
import wx
import weakref
import array
from scipy.misc import imresize
from ctypes import sizeof, c_float, c_void_p, c_uint, string_at

import matplotlib

from functools import wraps
from matplotlib.backends.backend_wx import FigureCanvasWx as Canvas
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as CanvasAgg
from matplotlib.backends.backend_wx import RendererWx
from ifigure.utils.cbook import EraseBitMap
from operator import itemgetter

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('BackendWXAggGL')

isMPL_before_1_2 = LooseVersion(matplotlib.__version__) < LooseVersion("1.2")

#
#  OpenGL extention
#
try:
    from wx import glcanvas
except ImportError:
    pass
try:
    # The Python OpenGL package can be found at
    # http://PyOpenGL.sourceforge.net/
    from OpenGL.GL import *
    from OpenGL.GLUT import *
    from OpenGL.GLU import *
    from OpenGL.GL import shaders
    from OpenGL.arrays import vbo

    class myvbo(vbo.VBO):
        pass
    haveOpenGL = True
except ImportError:
    haveOpenGL = False

near_clipping = 8.

multisample = 1
multisample_init_done = False

basedir = os.path.dirname(__file__)


def compile_file(file, mode):
    fid = open(os.path.join(basedir, file), 'r')
    prog = ''.join(fid.readlines())
    pl = shaders.compileShader(prog, mode)
    return pl


def get_vbo(data, *args, **kwargs):
    vbo = myvbo(data, *args, **kwargs)
    vbo.need_update = False
    return vbo


def read_glmatrix(mode):
    #    a = (GLfloat * 16)()
    return np.transpose(glGetFloatv(mode))
#    return np.transpuse(np.array(list(a)).reshape(4,4))


def define_unform(shader, name):
    shader.uniform_loc[name] = glGetUniformLocation(shader, name)


def check_framebuffer(message, mode=GL_FRAMEBUFFER):
    if (glCheckFramebufferStatus(mode) !=
            GL_FRAMEBUFFER_COMPLETE):
        print('Framebuffer imcomplete (' + message + ')')
        print(str(glCheckFramebufferStatus(GL_FRAMEBUFFER)))
        print(str(GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT))
        print(str(GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS))
        print(str(GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT))
        print(str(GL_FRAMEBUFFER_UNSUPPORTED))
        return False
    return True


def frustum(left, right, bottom, top, zNear, zFar):
    dx = right - left
    dy = top - bottom
    A = (right + left) / (right - left)
    B = (top + bottom) / (top - bottom)
    C = -(zFar + zNear) / (zFar - zNear)
    D = - (2*zFar * zNear) / (zFar - zNear)
    return np.array([[2*zNear/dx, 0,           A, 0],
                     [0,          2*zNear/dy,  B, 0],
                     [0,          0,           C, D],
                     [0,          0,          -1, 0]])


def ortho(left, right, bottom, top, zNear, zFar):
    dx = right - left
    dy = top - bottom
    dz = zFar - zNear
    tx = - (right + left) / (right - left)
    ty = - (top + bottom) / (top - bottom)
    tz = - (zFar + zNear) / (zFar - zNear)
    return np.array([[2/dx, 0,     0,     tx],
                     [0,    2/dy,  0,     ty],
                     [0,    0,     -2/dz, tz],
                     [0,    0,     0,     1.]])


def wait_gl_finish(method):
    @wraps(method)
    def method2(self, *args, **kargs):
        method(self, *args, **kargs)
        glFinish()
    return method


def check_gl_error():
    error = glGetError()
    if error != 0:
        print(("GL error ", error))


class vbos_dict(dict):
    def __del__(self, *args, **kwargs):
        if 'im' in self:
            if self['im'] is not None:
                dprint2('deleteing texture', self['im'])
                glDeleteTextures(self['im'])
                self['im'] = None
        return


#from renderer_gl import RendererGL


class RendererGLMixin(object):
    def __init__(self, *args, **kwargs):
        self.do_stencil_test = False
        self._no_update_id = False
        self._num_globj = 0

    def __del__(self):
        self._glcanvas = None

#    def set_glcanvas(self, glcanvas):
#        self._glcanvas = glcanvas

    def get_glcanvas(self):
        from ifigure.matplotlib_mod.backend_wxagg_gl import FigureCanvasWxAggModGL
        return FigureCanvasWxAggModGL.glcanvas

    def gl_draw_image(self, gc, path, trans, im, **kwargs):
        self.get_glcanvas().store_draw_request('image', gc,
                                               path, trans,
                                               im, **kwargs)

    def gl_draw_markers(self, gc, marker_path, marker_trans,
                        path, transform, rgbFace=None, **kwargs):
        self.get_glcanvas().store_draw_request('markers', gc, marker_path, marker_trans,
                                               path, rgbFace, **kwargs)

    def gl_draw_path(self, gc, path, transform, rgbFace=None, **kwargs):
        self.get_glcanvas().store_draw_request('path', gc, path, rgbFace, **kwargs)

    def gl_draw_path_collection(self, gc, frozen, paths,
                                transform, offset, transOffset,
                                facecolor, edgecolor,
                                linewidth, linestyle,
                                antialias, urls, offset_position, **kwargs):

        self.get_glcanvas().store_draw_request('path_collection', gc, paths,
                                               facecolor, edgecolor,
                                               linewidth, linestyle, offset,
                                               **kwargs)

    def gl_draw_path_collection_e(self, gc, frozen, paths,
                                  transform, offset, transOffset,
                                  facecolor, edgecolor,
                                  linewidth, linestyle,
                                  antialias, urls, offset_position, **kwargs):
        self.get_glcanvas().store_draw_request('path_collection_e', gc, paths,
                                               facecolor, edgecolor,
                                               linewidth, linestyle, offset,
                                               **kwargs)

#    def set_canvas_proj(self, worldM, viewM, perspM, lookat):
#        self.get_glcanvas().set_proj(worldM, viewM, perspM, lookat)

    def update_id_data(self, data, tag=None):
        if not self._no_update_id:
            tag._gl_id_data = data
        self._no_update_id = False

    def no_update_id(self):
        self._no_update_id = True


def mixin_gl_renderer(renderer):
    if hasattr(renderer, '_gl_renderer'):
        return
    renderer._gl_renderer = RendererGLMixin()
    renderer.gl_svg_rescale = False
    names = ('gl_draw_markers', 'gl_draw_path_collection',
             'gl_draw_path_collection_e',
             'gl_draw_path', 'gl_draw_image',
             'get_glcanvas',
             'update_id_data', 'no_update_id')
    for name in names:
        setattr(renderer, name, getattr(renderer._gl_renderer, name))


class FigureCanvasWxAggModGL(FigureCanvasWxAggMod):
    glcanvas = None

    def __init__(self, *args, **kwargs):
        FigureCanvasWxAggMod.__init__(self, *args, **kwargs)
        if FigureCanvasWxAggModGL.glcanvas is None:
            win = wx.GetApp().TopWindow
            glcanvas = MyGLCanvas(win)
            FigureCanvasWxAggModGL.glcanvas = glcanvas
            glcanvas.SetMinSize((2, 2))
            glcanvas.SetMaxSize((2, 2))
            # self.SetSizer(wx.BoxSizer(wx.HORIZONTAL))
            win.GetSizer().Add(glcanvas)
            win.Layout()
            glcanvas.Refresh()

    def get_renderer(self, cleared=False):
        l, b, w, h = self.figure.bbox.bounds
        key = w, h, self.figure.dpi
        try:
            self._lastKey, self.renderer
        except AttributeError:
            need_new_renderer = True
        else:
            need_new_renderer = (self._lastKey != key)

        if need_new_renderer:
            self.renderer = RendererAgg(w, h, self.figure.dpi)
            mixin_gl_renderer(self.renderer)
            self._lastKey = key
        elif cleared:
            self.renderer.clear()
        return self.renderer

    def draw(self, *args, **kargs):
        self._update_hl_color()
        return FigureCanvasWxAggMod.draw(self,  *args, **kargs)

    def draw_artist(self, drawDC=None, alist=None):
        if alist is None:
            alist = []
        gl_obj = [a for a in alist if hasattr(a, 'is_gl')]

        for o in gl_obj:
            o.is_last = False
        if len(gl_obj) > 0:
            gl_obj[-1].is_last = True
            self.renderer._k_globj = 0
            self.renderer._num_globj = len(gl_obj)
            self.renderer.no_update_id()
#            self.renderer.no_lighting = no_lighting
            self._update_hl_color()
            FigureCanvasWxAggModGL.glcanvas._artist_mask = alist

        v = FigureCanvasWxAggMod.draw_artist(self, drawDC=drawDC, alist=alist)
#        self.renderer.no_lighting = False
        return v

    def _update_hl_color(self):
        value = self.hl_color
        vv = list([float(x) for x in value]) + [1.0]
        vv[3] = 0.65
        FigureCanvasWxAggModGL.glcanvas._hl_color = tuple(vv[:4])

    def _onPaint(self, evt):
        #        self.glcanvas.OnPaint(evt)
        #        evt.Skip()
        FigureCanvasWxAggMod._onPaint(self, evt)

    def _onSize(self, evt=None, nocheck=False):
        FigureCanvasWxAggMod._onSize(self, evt=evt, nocheck=nocheck)
        # self.glcanvas.SetSize(self.bitmap.GetSize())

    def disable_alpha_blend(self):
        FigureCanvasWxAggModGL.glcanvas._alpha_blend = False

    def enable_alpha_blend(self):
        FigureCanvasWxAggModGL.glcanvas._alpha_blend = True
